#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.
import logging
from typing import Any, Dict, List, Callable, Optional

import numpy as np
import sklearn

from miplearn.components.primal import (
    _extract_bin_var_names_values,
    _extract_bin_var_names,
)
from miplearn.components.primal.actions import PrimalComponentAction
from miplearn.extractors.abstract import FeaturesExtractor
from miplearn.solvers.abstract import AbstractModel
from miplearn.h5 import H5File

logger = logging.getLogger(__name__)


class IndependentVarsPrimalComponent:
    def __init__(
        self,
        base_clf: Any,
        extractor: FeaturesExtractor,
        action: PrimalComponentAction,
        clone_fn: Callable[[Any], Any] = sklearn.clone,
    ):
        self.base_clf = base_clf
        self.extractor = extractor
        self.clf_: Dict[bytes, Any] = {}
        self.bin_var_names_: Optional[np.ndarray] = None
        self.n_features_: Optional[int] = None
        self.clone_fn = clone_fn
        self.action = action

    def fit(self, train_h5: List[str]) -> None:
        logger.info("Reading training data...")
        self.bin_var_names_ = None
        n_bin_vars: Optional[int] = None
        n_vars: Optional[int] = None
        x, y = [], []
        for h5_filename in train_h5:
            with H5File(h5_filename, "r") as h5:
                # Get number of variables
                var_types = h5.get_array("static_var_types")
                assert var_types is not None
                n_vars = len(var_types)

                # Extract features
                (
                    bin_var_names,
                    bin_var_values,
                    bin_var_indices,
                ) = _extract_bin_var_names_values(h5)

                # Store/check variable names
                if self.bin_var_names_ is None:
                    self.bin_var_names_ = bin_var_names
                    n_bin_vars = len(self.bin_var_names_)
                else:
                    assert np.all(bin_var_names == self.bin_var_names_)

                # Build x and y vectors
                x_sample = self.extractor.get_var_features(h5)
                assert len(x_sample.shape) == 2
                assert x_sample.shape[0] == n_vars
                x_sample = x_sample[bin_var_indices]
                if self.n_features_ is None:
                    self.n_features_ = x_sample.shape[1]
                else:
                    assert x_sample.shape[1] == self.n_features_
                x.append(x_sample)
                y.append(bin_var_values)

        assert n_bin_vars is not None
        assert self.bin_var_names_ is not None

        logger.info("Constructing matrices...")
        x_np = np.vstack(x)
        y_np = np.hstack(y)
        n_samples = len(train_h5) * n_bin_vars
        assert x_np.shape == (n_samples, self.n_features_)
        assert y_np.shape == (n_samples,)
        logger.info(
            f"Dataset has {n_bin_vars} binary variables, "
            f"{len(train_h5):,d} samples per variable, "
            f"{self.n_features_:,d} features, 1 target and 2 classes"
        )

        logger.info(f"Training {n_bin_vars} classifiers...")
        self.clf_ = {}
        for (var_idx, var_name) in enumerate(self.bin_var_names_):
            self.clf_[var_name] = self.clone_fn(self.base_clf)
            self.clf_[var_name].fit(
                x_np[var_idx::n_bin_vars, :], y_np[var_idx::n_bin_vars]
            )

        logger.info("Done fitting.")

    def before_mip(
        self, test_h5: str, model: AbstractModel, stats: Dict[str, Any]
    ) -> None:
        assert self.bin_var_names_ is not None
        assert self.n_features_ is not None

        # Read features
        with H5File(test_h5, "r") as h5:
            x_sample = self.extractor.get_var_features(h5)
            bin_var_names, bin_var_indices = _extract_bin_var_names(h5)
            assert np.all(bin_var_names == self.bin_var_names_)
            x_sample = x_sample[bin_var_indices]

        assert x_sample.shape == (len(self.bin_var_names_), self.n_features_)

        # Predict optimal solution
        logger.info("Predicting warm starts...")
        y_pred = []
        for (var_idx, var_name) in enumerate(self.bin_var_names_):
            x_var = x_sample[var_idx, :].reshape(1, -1)
            y_var = self.clf_[var_name].predict(x_var)
            assert y_var.shape == (1,)
            y_pred.append(y_var[0])

        # Construct warm starts, based on prediction
        y_pred_np = np.array(y_pred).reshape(1, -1)
        assert y_pred_np.shape == (1, len(self.bin_var_names_))
        self.action.perform(model, self.bin_var_names_, y_pred_np, stats)



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.
from os.path import exists
from typing import List, Any

from miplearn.h5 import H5File
from miplearn.io import load


class LearningSolver:
    def __init__(self, components: List[Any], skip_lp=False):
        self.components = components
        self.skip_lp = skip_lp

    def fit(self, train_filenames):
        train_h5 = [f.replace(".pkl.gz", ".h5") for f in train_filenames]
        for comp in self.components:
            comp.fit(train_h5)

    def optimize(self, data_filename, build_model):
        stats = {}
        h5_filename = data_filename.replace(".pkl.gz", ".h5")
        mode = "r+" if exists(h5_filename) else "w"
        with H5File(h5_filename, mode) as h5:
            model = load(data_filename, build_model)
            model.extract_after_load(h5)
            if not self.skip_lp:
                relaxed = model.relax()
                relaxed.optimize()
                relaxed.extract_after_lp(h5)
            for comp in self.components:
                comp.before_mip(h5_filename, model, stats)
            model.optimize()
            model.extract_after_mip(h5)

        model.optimize()
        return stats



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

import logging
from abc import ABC, abstractmethod
from typing import List, Dict, Any, Optional, Tuple

import numpy as np

from . import _extract_bin_var_names_values
from .actions import PrimalComponentAction
from ...extractors.abstract import FeaturesExtractor
from ...solvers.abstract import AbstractModel
from ...h5 import H5File

logger = logging.getLogger()


class SolutionConstructor(ABC):
    @abstractmethod
    def construct(self, y_proba: np.ndarray, solutions: np.ndarray) -> np.ndarray:
        pass


class MemorizingPrimalComponent:
    """
    Component that memorizes all solutions seen during training, then fits a
    single classifier to predict which of the memorized solutions should be
    provided to the solver. Optionally combines multiple memorized solutions
    into a single, partial one.
    """

    def __init__(
        self,
        clf: Any,
        extractor: FeaturesExtractor,
        constructor: SolutionConstructor,
        action: PrimalComponentAction,
    ) -> None:
        assert clf is not None
        self.clf = clf
        self.extractor = extractor
        self.constructor = constructor
        self.solutions_: Optional[np.ndarray] = None
        self.bin_var_names_: Optional[np.ndarray] = None
        self.action = action

    def fit(self, train_h5: List[str]) -> None:
        logger.info("Reading training data...")
        n_samples = len(train_h5)
        solutions_ = []
        self.bin_var_names_ = None
        x, y, n_features = [], [], None
        solution_to_idx: Dict[Tuple, int] = {}
        for h5_filename in train_h5:
            with H5File(h5_filename, "r") as h5:
                bin_var_names, bin_var_values, _ = _extract_bin_var_names_values(h5)

                # Store/check variable names
                if self.bin_var_names_ is None:
                    self.bin_var_names_ = bin_var_names
                else:
                    assert np.all(bin_var_names == self.bin_var_names_)

                # Store solution
                sol = tuple(np.where(bin_var_values)[0])
                if sol not in solution_to_idx:
                    solutions_.append(bin_var_values)
                    solution_to_idx[sol] = len(solution_to_idx)
                y.append(solution_to_idx[sol])

                # Extract features
                x_sample = self.extractor.get_instance_features(h5)
                assert len(x_sample.shape) == 1
                if n_features is None:
                    n_features = len(x_sample)
                else:
                    assert len(x_sample) == n_features
                x.append(x_sample)

        logger.info("Constructing matrices...")
        x_np = np.vstack(x)
        y_np = np.array(y)
        assert len(x_np.shape) == 2
        assert x_np.shape[0] == n_samples
        assert x_np.shape[1] == n_features
        assert y_np.shape == (n_samples,)
        self.solutions_ = np.array(solutions_)
        n_classes = len(solution_to_idx)
        logger.info(
            f"Dataset has {n_samples:,d} samples, "
            f"{n_features:,d} features and {n_classes:,d} classes"
        )

        logger.info("Training classifier...")
        self.clf.fit(x_np, y_np)

        logger.info("Done fitting.")

    def before_mip(
        self, test_h5: str, model: AbstractModel, stats: Dict[str, Any]
    ) -> None:
        assert self.solutions_ is not None
        assert self.bin_var_names_ is not None

        # Read features
        with H5File(test_h5, "r") as h5:
            x_sample = self.extractor.get_instance_features(h5)
        assert len(x_sample.shape) == 1
        x_sample = x_sample.reshape(1, -1)

        # Predict optimal solution
        logger.info("Predicting primal solution...")
        y_proba = self.clf.predict_proba(x_sample)
        assert len(y_proba.shape) == 2
        assert y_proba.shape[0] == 1
        assert y_proba.shape[1] == len(self.solutions_)

        # Construct warm starts, based on prediction
        starts = self.constructor.construct(y_proba[0, :], self.solutions_)
        self.action.perform(model, self.bin_var_names_, starts, stats)


class SelectTopSolutions(SolutionConstructor):
    """
    Warm start construction strategy that selects and returns the top k solutions.
    """

    def __init__(self, k: int) -> None:
        self.k = k

    def construct(self, y_proba: np.ndarray, solutions: np.ndarray) -> np.ndarray:
        # Check arguments
        assert len(y_proba.shape) == 1
        assert len(solutions.shape) == 2
        assert len(y_proba) == solutions.shape[0]

        # Select top k solutions
        ind = np.argsort(-y_proba, kind="stable")
        selected = ind[: min(self.k, len(ind))]
        return solutions[selected, :]


class MergeTopSolutions(SolutionConstructor):
    """
    Warm start construction strategy that first selects the top k solutions,
    then merges them into a single solution.

    To merge the solutions, the strategy first computes the mean optimal value of each
    decision variable, then: (i) sets the variable to zero if the mean is below
    thresholds[0]; (ii) sets the variable to one if the mean is above thresholds[1];
    (iii) leaves the variable free otherwise.
    """

    def __init__(self, k: int, thresholds: List[float]):
        assert len(thresholds) == 2
        self.k = k
        self.thresholds = thresholds

    def construct(self, y_proba: np.ndarray, solutions: np.ndarray) -> np.ndarray:
        filtered = SelectTopSolutions(self.k).construct(y_proba, solutions)
        mean = filtered.mean(axis=0)
        start = np.full((1, solutions.shape[1]), float("nan"))
        start[0, mean <= self.thresholds[0]] = 0
        start[0, mean >= self.thresholds[1]] = 1
        return start



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.
from typing import List, Dict, Any
from unittest.mock import Mock, call

from sklearn.dummy import DummyClassifier

from miplearn.components.primal.actions import SetWarmStart
from miplearn.components.primal.indep import IndependentVarsPrimalComponent
from miplearn.extractors.fields import H5FieldsExtractor


def test_indep(multiknapsack_h5: List[str]) -> None:
    # Create and fit component
    clone_fn = Mock(return_value=Mock(wraps=DummyClassifier()))
    comp = IndependentVarsPrimalComponent(
        base_clf="dummy",
        extractor=H5FieldsExtractor(var_fields=["lp_var_values"]),
        clone_fn=clone_fn,
        action=SetWarmStart(),
    )
    comp.fit(multiknapsack_h5)

    # Should call clone 100 times and store the 100 classifiers
    clone_fn.assert_has_calls([call("dummy") for _ in range(100)])
    assert len(comp.clf_) == 100

    for v in [b"x[0]", b"x[1]"]:
        # Should pass correct data to fit
        comp.clf_[v].fit.assert_called()
        x, y = comp.clf_[v].fit.call_args.args
        assert x.shape == (3, 1)
        assert y.shape == (3,)

    # Call before-mip
    stats: Dict[str, Any] = {}
    model = Mock()
    comp.before_mip(multiknapsack_h5[0], model, stats)

    # Should call predict with correct args
    for v in [b"x[0]", b"x[1]"]:
        comp.clf_[v].predict.assert_called()
        (x_test,) = comp.clf_[v].predict.call_args.args
        assert x_test.shape == (1, 1)

    # Should set warm starts
    model.set_warm_starts.assert_called()
    names, starts, _ = model.set_warm_starts.call_args.args
    assert len(names) == 100
    assert starts.shape == (1, 100)



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

from types import TracebackType
from typing import Optional, Any, Union, List, Type, Literal

import h5py
import numpy as np
from scipy.sparse import coo_matrix

Bytes = Union[bytes, bytearray]

Scalar = Union[None, bool, str, int, float]

Vector = Union[
    None,
    List[bool],
    List[str],
    List[int],
    List[float],
    List[Optional[str]],
    np.ndarray,
]

VectorList = Union[
    List[List[bool]],
    List[List[str]],
    List[List[int]],
    List[List[float]],
    List[Optional[List[bool]]],
    List[Optional[List[str]]],
    List[Optional[List[int]]],
    List[Optional[List[float]]],
]


class H5File:
    def __init__(
        self,
        filename: str,
        mode: str = "r+",
    ) -> None:
        self.file = h5py.File(filename, mode, libver="latest")

    def get_scalar(self, key: str) -> Optional[Any]:
        if key not in self.file:
            return None
        ds = self.file[key]
        assert (
            len(ds.shape) == 0
        ), f"0-dimensional array expected; found shape {ds.shape}"
        if h5py.check_string_dtype(ds.dtype):
            return ds.asstr()[()]
        else:
            return ds[()].tolist()

    def put_scalar(self, key: str, value: Any) -> None:
        if value is None:
            return
        self._assert_is_scalar(value)
        if key in self.file:
            del self.file[key]
        self.file.create_dataset(key, data=value)

    def put_array(self, key: str, value: Optional[np.ndarray]) -> None:
        if value is None:
            return
        self._assert_is_array(value)
        if value.dtype.kind == "f":
            value = value.astype("float32")
        if key in self.file:
            del self.file[key]
        return self.file.create_dataset(key, data=value, compression="gzip")

    def get_array(self, key: str) -> Optional[np.ndarray]:
        if key not in self.file:
            return None
        return self.file[key][:]

    def put_sparse(self, key: str, value: coo_matrix) -> None:
        if value is None:
            return
        self._assert_is_sparse(value)
        self.put_array(f"{key}_row", value.row)
        self.put_array(f"{key}_col", value.col)
        self.put_array(f"{key}_data", value.data)

    def get_sparse(self, key: str) -> Optional[coo_matrix]:
        row = self.get_array(f"{key}_row")
        if row is None:
            return None
        col = self.get_array(f"{key}_col")
        data = self.get_array(f"{key}_data")
        assert col is not None
        assert data is not None
        return coo_matrix((data, (row, col)))

    def get_bytes(self, key: str) -> Optional[Bytes]:
        if key not in self.file:
            return None
        ds = self.file[key]
        assert (
            len(ds.shape) == 1
        ), f"1-dimensional array expected; found shape {ds.shape}"
        return ds[()].tobytes()

    def put_bytes(self, key: str, value: Bytes) -> None:
        assert isinstance(
            value, (bytes, bytearray)
        ), f"bytes expected; found: {value.__class__}"  # type: ignore
        self.put_array(key, np.frombuffer(value, dtype="uint8"))

    def __enter__(self) -> "H5File":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> Literal[False]:
        self.file.close()
        return False

    def _assert_is_scalar(self, value: Any) -> None:
        if value is None:
            return
        if isinstance(value, (str, bool, int, float, bytes, np.bytes_)):
            return
        assert False, f"scalar expected; found instead: {value} ({value.__class__})"

    def _assert_is_array(self, value: np.ndarray) -> None:
        assert isinstance(
            value, np.ndarray
        ), f"np.ndarray expected; found instead: {value.__class__}"
        assert value.dtype.kind in "biufS", f"Unsupported dtype: {value.dtype}"

    def _assert_is_sparse(self, value: Any) -> None:
        assert isinstance(
            value, coo_matrix
        ), f"coo_matrix expected; found: {value.__class__}"
        self._assert_is_array(value.data)



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.
from typing import List, Dict, Any
from unittest.mock import Mock

from sklearn.dummy import DummyClassifier

from miplearn.components.primal.actions import SetWarmStart
from miplearn.components.primal.joint import JointVarsPrimalComponent
from miplearn.extractors.fields import H5FieldsExtractor


def test_joint(multiknapsack_h5: List[str]) -> None:
    # Create mock classifier
    clf = Mock(wraps=DummyClassifier())

    # Create and fit component
    comp = JointVarsPrimalComponent(
        clf=clf,
        extractor=H5FieldsExtractor(instance_fields=["static_var_obj_coeffs"]),
        action=SetWarmStart(),
    )
    comp.fit(multiknapsack_h5)

    # Should call fit method with correct arguments
    clf.fit.assert_called()
    x, y = clf.fit.call_args.args
    assert x.shape == (3, 100)
    assert y.shape == (3, 100)

    # Call before-mip
    stats: Dict[str, Any] = {}
    model = Mock()
    comp.before_mip(multiknapsack_h5[0], model, stats)

    # Should call predict with correct args
    clf.predict.assert_called()
    (x_test,) = clf.predict.call_args.args
    assert x_test.shape == (1, 100)

    # Should set warm starts
    model.set_warm_starts.assert_called()
    names, starts, _ = model.set_warm_starts.call_args.args
    assert len(names) == 100
    assert starts.shape == (1, 100)



project = "MIPLearn"
copyright = "2020-2022, UChicago Argonne, LLC"
author = ""
release = "0.3"
extensions = [
    "myst_parser",
    "nbsphinx",
    "sphinx_multitoc_numbering",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
]
templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
html_theme = "sphinx_book_theme"
html_static_path = ["_static"]
html_css_files = [
    "custom.css",
]
html_theme_options = {
    "repository_url": "https://github.com/ANL-CEEESA/MIPLearn/",
    "use_repository_button": False,
    "extra_navbar": "",
}
html_title = f"MIPLearn {release}"
nbsphinx_execute = "never"



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

from dataclasses import dataclass
from typing import List

import gurobipy as gp
import networkx as nx
import numpy as np
import pyomo.environ as pe
from gurobipy import GRB, quicksum
from networkx import Graph
from scipy.stats import uniform, randint
from scipy.stats.distributions import rv_frozen

from miplearn.solvers.gurobi import GurobiModel
from miplearn.solvers.pyomo import PyomoModel


@dataclass
class MaxWeightStableSetData:
    graph: Graph
    weights: np.ndarray


class MaxWeightStableSetGenerator:
    """Random instance generator for the Maximum-Weight Stable Set Problem.

    The generator has two modes of operation. When `fix_graph=True` is provided,
    one random Erdős-Rényi graph $G_{n,p}$ is generated in the constructor, where $n$
    and $p$ are sampled from user-provided probability distributions `n` and `p`. To
    generate each instance, the generator independently samples each $w_v$ from the
    user-provided probability distribution `w`.

    When `fix_graph=False`, a new random graph is generated for each instance; the
    remaining parameters are sampled in the same way.
    """

    def __init__(
        self,
        w: rv_frozen = uniform(loc=10.0, scale=1.0),
        n: rv_frozen = randint(low=250, high=251),
        p: rv_frozen = uniform(loc=0.05, scale=0.0),
        fix_graph: bool = True,
    ):
        """Initialize the problem generator.

        Parameters
        ----------
        w: rv_continuous
            Probability distribution for vertex weights.
        n: rv_discrete
            Probability distribution for parameter $n$ in Erdős-Rényi model.
        p: rv_continuous
            Probability distribution for parameter $p$ in Erdős-Rényi model.
        """
        assert isinstance(w, rv_frozen), "w should be a SciPy probability distribution"
        assert isinstance(n, rv_frozen), "n should be a SciPy probability distribution"
        assert isinstance(p, rv_frozen), "p should be a SciPy probability distribution"
        self.w = w
        self.n = n
        self.p = p
        self.fix_graph = fix_graph
        self.graph = None
        if fix_graph:
            self.graph = self._generate_graph()

    def generate(self, n_samples: int) -> List[MaxWeightStableSetData]:
        def _sample() -> MaxWeightStableSetData:
            if self.graph is not None:
                graph = self.graph
            else:
                graph = self._generate_graph()
            weights = np.round(self.w.rvs(graph.number_of_nodes()), 2)
            return MaxWeightStableSetData(graph, weights)

        return [_sample() for _ in range(n_samples)]

    def _generate_graph(self) -> Graph:
        return nx.generators.random_graphs.binomial_graph(self.n.rvs(), self.p.rvs())


def build_stab_model_gurobipy(data: MaxWeightStableSetData) -> GurobiModel:
    model = gp.Model()
    nodes = list(data.graph.nodes)
    x = model.addVars(nodes, vtype=GRB.BINARY, name="x")
    model.setObjective(quicksum(-data.weights[i] * x[i] for i in nodes))
    for clique in nx.find_cliques(data.graph):
        model.addConstr(quicksum(x[i] for i in clique) <= 1)
    model.update()
    return GurobiModel(model)


def build_stab_model_pyomo(
    data: MaxWeightStableSetData,
    solver="gurobi_persistent",
) -> PyomoModel:
    model = pe.ConcreteModel()
    nodes = pe.Set(initialize=list(data.graph.nodes))
    model.x = pe.Var(nodes, domain=pe.Boolean, name="x")
    model.obj = pe.Objective(expr=sum([-data.weights[i] * model.x[i] for i in nodes]))
    model.clique_eqs = pe.ConstraintList()
    for clique in nx.find_cliques(data.graph):
        model.clique_eqs.add(expr=sum(model.x[i] for i in clique) <= 1)
    return PyomoModel(model, solver)



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

import numpy as np
from scipy.stats import uniform, randint

from miplearn.problems.pmedian import PMedianGenerator, build_pmedian_model


def test_pmedian() -> None:
    np.random.seed(42)
    gen = PMedianGenerator(
        x=uniform(loc=0.0, scale=100.0),
        y=uniform(loc=0.0, scale=100.0),
        n=randint(low=5, high=6),
        p=randint(low=2, high=3),
        demands=uniform(loc=0, scale=20),
        capacities=uniform(loc=0, scale=100),
        distances_jitter=uniform(loc=0.95, scale=0.1),
        demands_jitter=uniform(loc=0.95, scale=0.1),
        capacities_jitter=uniform(loc=0.95, scale=0.1),
        fixed=True,
    )
    data = gen.generate(2)

    assert data[0].p == 2
    assert data[0].demands.tolist() == [0.41, 19.4, 16.65, 4.25, 3.64]
    assert data[0].capacities.tolist() == [18.34, 30.42, 52.48, 43.19, 29.12]
    assert data[0].distances.tolist() == [
        [0.0, 50.17, 82.42, 32.76, 33.2],
        [50.17, 0.0, 72.64, 72.51, 17.06],
        [82.42, 72.64, 0.0, 71.69, 70.92],
        [32.76, 72.51, 71.69, 0.0, 56.56],
        [33.2, 17.06, 70.92, 56.56, 0.0],
    ]

    assert data[1].p == 2
    assert data[1].demands.tolist() == [0.42, 19.03, 16.68, 4.27, 3.53]
    assert data[1].capacities.tolist() == [19.2, 31.26, 54.79, 44.9, 29.41]
    assert data[1].distances.tolist() == [
        [0.0, 51.6, 83.31, 33.77, 31.95],
        [51.6, 0.0, 70.25, 71.09, 17.05],
        [83.31, 70.25, 0.0, 68.81, 67.62],
        [33.77, 71.09, 68.81, 0.0, 58.88],
        [31.95, 17.05, 67.62, 58.88, 0.0],
    ]

    model = build_pmedian_model(data[0])
    assert model.inner.numVars == 30
    assert model.inner.numConstrs == 11
    model.optimize()
    assert round(model.inner.objVal) == 107



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

import numpy as np
from scipy.stats import uniform, randint

from miplearn.problems.uc import (
    UnitCommitmentData,
    build_uc_model,
    UnitCommitmentGenerator,
)


def test_generator() -> None:
    np.random.seed(42)
    gen = UnitCommitmentGenerator(
        n_units=randint(low=3, high=4),
        n_periods=randint(low=4, high=5),
        max_power=uniform(loc=50, scale=450),
        min_power=uniform(loc=0.25, scale=0.5),
        cost_startup=uniform(loc=1, scale=1),
        cost_prod=uniform(loc=1, scale=1),
        cost_fixed=uniform(loc=1, scale=1),
        min_uptime=randint(low=1, high=8),
        min_downtime=randint(low=1, high=8),
        cost_jitter=uniform(loc=0.75, scale=0.5),
        demand_jitter=uniform(loc=0.9, scale=0.2),
        fix_units=True,
    )
    data = gen.generate(2)

    assert data[0].demand.tolist() == [430.3, 518.65, 448.16, 860.61]
    assert data[0].min_power.tolist() == [120.05, 156.73, 124.44]
    assert data[0].max_power.tolist() == [218.54, 477.82, 379.4]
    assert data[0].min_uptime.tolist() == [3, 3, 5]
    assert data[0].min_downtime.tolist() == [4, 3, 6]
    assert data[0].cost_startup.tolist() == [1.06, 1.72, 1.94]
    assert data[0].cost_prod.tolist() == [1.0, 1.99, 1.62]
    assert data[0].cost_fixed.tolist() == [1.61, 1.01, 1.02]

    assert data[1].demand.tolist() == [407.3, 476.18, 458.77, 840.38]
    assert data[1].min_power.tolist() == [120.05, 156.73, 124.44]
    assert data[1].max_power.tolist() == [218.54, 477.82, 379.4]
    assert data[1].min_uptime.tolist() == [3, 3, 5]
    assert data[1].min_downtime.tolist() == [4, 3, 6]
    assert data[1].cost_startup.tolist() == [1.32, 1.69, 2.29]
    assert data[1].cost_prod.tolist() == [1.09, 1.94, 1.23]
    assert data[1].cost_fixed.tolist() == [1.97, 1.04, 0.96]


def test_uc() -> None:
    data = UnitCommitmentData(
        demand=np.array([10, 12, 15, 10, 8, 5]),
        min_power=np.array([5, 5, 10]),
        max_power=np.array([10, 8, 20]),
        min_uptime=np.array([4, 3, 2]),
        min_downtime=np.array([4, 3, 2]),
        cost_startup=np.array([100, 120, 200]),
        cost_prod=np.array([1.0, 1.25, 1.5]),
        cost_fixed=np.array([10, 12, 9]),
    )
    model = build_uc_model(data)
    model.optimize()
    assert model.inner.objVal == 154.5


if __name__ == "__main__":
    data = UnitCommitmentGenerator().generate(1)[0]
    model = build_uc_model(data)
    model.optimize()



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

from glob import glob
from os.path import dirname
from typing import List

import pytest

from miplearn.extractors.fields import H5FieldsExtractor
from miplearn.extractors.abstract import FeaturesExtractor


@pytest.fixture()
def multiknapsack_h5() -> List[str]:
    return sorted(glob(f"{dirname(__file__)}/fixtures/multiknapsack*.h5"))


@pytest.fixture()
def default_extractor() -> FeaturesExtractor:
    return H5FieldsExtractor(
        instance_fields=["static_var_obj_coeffs"],
        var_fields=["lp_var_features"],
    )



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

import json
import os
from io import StringIO
from os.path import exists
from typing import Callable, List
from math import isfinite

from ..h5 import H5File
from ..io import _RedirectOutput, load, gzip
from ..parallel import p_umap


class BasicCollector:
    def collect(
        self,
        filenames: List[str],
        build_model: Callable,
        n_jobs: int = 1,
        progress: bool = False,
    ) -> None:
        def _collect(filename):
            h5_filename = filename.replace(".pkl.gz", ".h5")
            mps_filename = filename.replace(".pkl.gz", ".mps")

            if exists(h5_filename):
                # Try to read optimal solution
                mip_var_values = None
                try:
                    with H5File(h5_filename, "r") as h5:
                        mip_var_values = h5.get_array("mip_var_values")
                except:
                    pass

                if mip_var_values is None:
                    print(f"Removing empty/corrupted h5 file: {h5_filename}")
                    os.remove(h5_filename)
                else:
                    return

            with H5File(h5_filename, "w") as h5:
                streams = [StringIO()]
                with _RedirectOutput(streams):
                    # Load and extract static features
                    model = load(filename, build_model)
                    model.extract_after_load(h5)

                    # Solve LP relaxation
                    relaxed = model.relax()
                    relaxed.optimize()
                    relaxed.extract_after_lp(h5)

                    # Solve MIP
                    model.optimize()
                    model.extract_after_mip(h5)

                    # Add lazy constraints to model
                    if (
                        hasattr(model, "fix_violations")
                        and model.fix_violations is not None
                    ):
                        model.fix_violations(model, model.violations_, "aot")
                        h5.put_scalar(
                            "mip_constr_violations", json.dumps(model.violations_)
                        )

                    # Save MPS file
                    model.write(mps_filename)
                    gzip(mps_filename)

                h5.put_scalar("mip_log", streams[0].getvalue())

        if n_jobs > 1:
            p_umap(
                _collect,
                filenames,
                num_cpus=n_jobs,
                desc="collect",
                smoothing=0,
                disable=not progress,
            )
        else:
            for filename in filenames:
                _collect(filename)



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

import logging
from typing import Any, Dict, List

from . import _extract_bin_var_names_values
from .actions import PrimalComponentAction
from ...solvers.abstract import AbstractModel
from ...h5 import H5File

logger = logging.getLogger(__name__)


class ExpertPrimalComponent:
    def __init__(self, action: PrimalComponentAction):
        self.action = action

    """
    Component that predicts warm starts by peeking at the optimal solution.
    """

    def fit(self, train_h5: List[str]) -> None:
        pass

    def before_mip(
        self, test_h5: str, model: AbstractModel, stats: Dict[str, Any]
    ) -> None:
        with H5File(test_h5, "r") as h5:
            names, values, _ = _extract_bin_var_names_values(h5)
            self.action.perform(model, names, values.reshape(1, -1), stats)



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

import os
import subprocess
from typing import Callable

from ..h5 import H5File


class BranchPriorityCollector:
    def __init__(
        self,
        time_limit: float = 900.0,
        print_interval: int = 1,
        node_limit: int = 500,
    ) -> None:
        self.time_limit = time_limit
        self.print_interval = print_interval
        self.node_limit = node_limit

    def collect(self, data: str, _: Callable) -> None:
        basename = data.replace(".pkl.gz", "")
        env = os.environ.copy()
        env["JULIA_NUM_THREADS"] = "1"
        ret = subprocess.run(
            [
                "julia",
                "--project=.",
                "-e",
                (
                    f"using CPLEX, JuMP, MIPLearn.BB; "
                    f"BB.solve!("
                    f'    optimizer_with_attributes(CPLEX.Optimizer, "CPXPARAM_Threads" => 1),'
                    f'    "{basename}",'
                    f"    print_interval={self.print_interval},"
                    f"    time_limit={self.time_limit:.2f},"
                    f"    node_limit={self.node_limit},"
                    f")"
                ),
            ],
            check=True,
            capture_output=True,
            env=env,
        )
        h5_filename = f"{basename}.h5"
        with H5File(h5_filename, "r+") as h5:
            h5.put_scalar("bb_log", ret.stdout)



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.
import logging
from abc import ABC, abstractmethod
from typing import Optional, Dict

import numpy as np

from miplearn.solvers.abstract import AbstractModel

logger = logging.getLogger()


class PrimalComponentAction(ABC):
    @abstractmethod
    def perform(
        self,
        model: AbstractModel,
        var_names: np.ndarray,
        var_values: np.ndarray,
        stats: Optional[Dict],
    ) -> None:
        pass


class SetWarmStart(PrimalComponentAction):
    def perform(
        self,
        model: AbstractModel,
        var_names: np.ndarray,
        var_values: np.ndarray,
        stats: Optional[Dict],
    ) -> None:
        logger.info("Setting warm starts...")
        model.set_warm_starts(var_names, var_values, stats)


class FixVariables(PrimalComponentAction):
    def perform(
        self,
        model: AbstractModel,
        var_names: np.ndarray,
        var_values: np.ndarray,
        stats: Optional[Dict],
    ) -> None:
        logger.info("Fixing variables...")
        assert len(var_values.shape) == 2
        assert var_values.shape[0] == 1
        var_values = var_values.reshape(-1)
        model.fix_variables(var_names, var_values, stats)
        if stats is not None:
            stats["Heuristic"] = True


class EnforceProximity(PrimalComponentAction):
    def __init__(self, tol: float) -> None:
        self.tol = tol

    def perform(
        self,
        model: AbstractModel,
        var_names: np.ndarray,
        var_values: np.ndarray,
        stats: Optional[Dict],
    ) -> None:
        assert len(var_values.shape) == 2
        assert var_values.shape[0] == 1
        var_values = var_values.reshape(-1)

        constr_lhs = []
        constr_vars = []
        constr_rhs = 0.0
        for (i, var_name) in enumerate(var_names):
            if np.isnan(var_values[i]):
                continue
            constr_lhs.append(1.0 if var_values[i] < 0.5 else -1.0)
            constr_rhs -= var_values[i]
            constr_vars.append(var_name)

        constr_rhs += len(constr_vars) * self.tol
        logger.info(
            f"Adding proximity constraint (tol={self.tol}, nz={len(constr_vars)})..."
        )

        model.add_constrs(
            var_names=np.array(constr_vars),
            constrs_lhs=np.array([constr_lhs]),
            constrs_sense=np.array(["<"], dtype="S"),
            constrs_rhs=np.array([constr_rhs]),
        )
        if stats is not None:
            stats["Heuristic"] = True



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.
import logging
from typing import List, Dict, Any, Optional

import numpy as np

from miplearn.components.primal import _extract_bin_var_names_values
from miplearn.components.primal.actions import PrimalComponentAction
from miplearn.extractors.abstract import FeaturesExtractor
from miplearn.solvers.abstract import AbstractModel
from miplearn.h5 import H5File

logger = logging.getLogger(__name__)


class JointVarsPrimalComponent:
    def __init__(
        self, clf: Any, extractor: FeaturesExtractor, action: PrimalComponentAction
    ):
        self.clf = clf
        self.extractor = extractor
        self.bin_var_names_: Optional[np.ndarray] = None
        self.action = action

    def fit(self, train_h5: List[str]) -> None:
        logger.info("Reading training data...")
        self.bin_var_names_ = None
        x, y, n_samples, n_features = [], [], len(train_h5), None
        for h5_filename in train_h5:
            with H5File(h5_filename, "r") as h5:
                bin_var_names, bin_var_values, _ = _extract_bin_var_names_values(h5)

                # Store/check variable names
                if self.bin_var_names_ is None:
                    self.bin_var_names_ = bin_var_names
                else:
                    assert np.all(bin_var_names == self.bin_var_names_)

                # Build x and y vectors
                x_sample = self.extractor.get_instance_features(h5)
                assert len(x_sample.shape) == 1
                if n_features is None:
                    n_features = len(x_sample)
                else:
                    assert len(x_sample) == n_features
                x.append(x_sample)
                y.append(bin_var_values)
        assert self.bin_var_names_ is not None

        logger.info("Constructing matrices...")
        x_np = np.vstack(x)
        y_np = np.array(y)
        assert len(x_np.shape) == 2
        assert x_np.shape[0] == n_samples
        assert x_np.shape[1] == n_features
        assert y_np.shape == (n_samples, len(self.bin_var_names_))
        logger.info(
            f"Dataset has {n_samples:,d} samples, "
            f"{n_features:,d} features and {y_np.shape[1]:,d} targets"
        )

        logger.info("Training classifier...")
        self.clf.fit(x_np, y_np)

        logger.info("Done fitting.")

    def before_mip(
        self, test_h5: str, model: AbstractModel, stats: Dict[str, Any]
    ) -> None:
        assert self.bin_var_names_ is not None

        # Read features
        with H5File(test_h5, "r") as h5:
            x_sample = self.extractor.get_instance_features(h5)
        assert len(x_sample.shape) == 1
        x_sample = x_sample.reshape(1, -1)

        # Predict optimal solution
        logger.info("Predicting warm starts...")
        y_pred = self.clf.predict(x_sample)
        assert len(y_pred.shape) == 2
        assert y_pred.shape[0] == 1
        assert y_pred.shape[1] == len(self.bin_var_names_)

        # Construct warm starts, based on prediction
        self.action.perform(model, self.bin_var_names_, y_pred, stats)



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

from typing import List

from miplearn.extractors.dummy import DummyExtractor
from miplearn.h5 import H5File


def test_dummy(multiknapsack_h5: List[str]) -> None:
    ext = DummyExtractor()
    with H5File(multiknapsack_h5[0], "r") as h5:
        x = ext.get_instance_features(h5)
        assert x.shape == (1,)
        x = ext.get_var_features(h5)
        assert x.shape == (100, 1)
        x = ext.get_constr_features(h5)
        assert x.shape == (4, 1)



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

from dataclasses import dataclass
from typing import List

import gurobipy as gp
import numpy as np
from gurobipy.gurobipy import GRB
from scipy.stats import uniform, randint
from scipy.stats.distributions import rv_frozen

from .setcover import SetCoverGenerator
from miplearn.solvers.gurobi import GurobiModel


@dataclass
class SetPackData:
    costs: np.ndarray
    incidence_matrix: np.ndarray


class SetPackGenerator:
    def __init__(
        self,
        n_elements: rv_frozen = randint(low=50, high=51),
        n_sets: rv_frozen = randint(low=100, high=101),
        costs: rv_frozen = uniform(loc=0.0, scale=100.0),
        costs_jitter: rv_frozen = uniform(loc=-5.0, scale=10.0),
        K: rv_frozen = uniform(loc=25.0, scale=0.0),
        density: rv_frozen = uniform(loc=0.02, scale=0.00),
        fix_sets: bool = True,
    ) -> None:
        self.gen = SetCoverGenerator(
            n_elements=n_elements,
            n_sets=n_sets,
            costs=costs,
            costs_jitter=costs_jitter,
            K=K,
            density=density,
            fix_sets=fix_sets,
        )

    def generate(self, n_samples: int) -> List[SetPackData]:
        return [
            SetPackData(
                s.costs,
                s.incidence_matrix,
            )
            for s in self.gen.generate(n_samples)
        ]


def build_setpack_model(data: SetPackData) -> GurobiModel:
    (n_elements, n_sets) = data.incidence_matrix.shape
    model = gp.Model()
    x = model.addMVar(n_sets, vtype=GRB.BINARY, name="x")
    model.addConstr(data.incidence_matrix @ x <= np.ones(n_elements))
    model.setObjective(-data.costs @ x)
    model.update()
    return GurobiModel(model)



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

from dataclasses import dataclass
from typing import List, Optional

import gurobipy as gp
import numpy as np
from gurobipy import GRB, quicksum
from scipy.stats import uniform, randint
from scipy.stats.distributions import rv_frozen

from miplearn.solvers.gurobi import GurobiModel


@dataclass
class BinPackData:
    """Data for the bin packing problem.

    Parameters
    ----------
    sizes
        Sizes of the items
    capacity
        Capacity of the bin
    """

    sizes: np.ndarray
    capacity: int


class BinPackGenerator:
    """Random instance generator for the bin packing problem.

    If `fix_items=False`, the class samples the user-provided probability distributions
    n, sizes and capacity to decide, respectively, the number of items, the sizes of
    the items and capacity of the bin. All values are sampled independently.

    If `fix_items=True`, the class creates a reference instance, using the method
    previously described, then generates additional instances by perturbing its item
    sizes and bin capacity. More specifically, the sizes of the items are set to `s_i
    * gamma_i` where `s_i` is the size of the i-th item in the reference instance and
    `gamma_i` is sampled from `sizes_jitter`. Similarly, the bin capacity is set to `B *
    beta`, where `B` is the reference bin capacity and `beta` is sampled from
    `capacity_jitter`. The number of items remains the same across all generated
    instances.

    Args
    ----
    n
        Probability distribution for the number of items.
    sizes
        Probability distribution for the item sizes.
    capacity
        Probability distribution for the bin capacity.
    sizes_jitter
        Probability distribution for the item size randomization.
    capacity_jitter
        Probability distribution for the bin capacity.
    fix_items
        If `True`, generates a reference instance, then applies some perturbation to it.
        If `False`, generates completely different instances.
    """

    def __init__(
        self,
        n: rv_frozen,
        sizes: rv_frozen,
        capacity: rv_frozen,
        sizes_jitter: rv_frozen,
        capacity_jitter: rv_frozen,
        fix_items: bool,
    ) -> None:
        self.n = n
        self.sizes = sizes
        self.capacity = capacity
        self.sizes_jitter = sizes_jitter
        self.capacity_jitter = capacity_jitter
        self.fix_items = fix_items
        self.ref_data: Optional[BinPackData] = None

    def generate(self, n_samples: int) -> List[BinPackData]:
        """Generates random instances.

        Parameters
        ----------
        n_samples
            Number of samples to generate.
        """

        def _sample() -> BinPackData:
            if self.ref_data is None:
                n = self.n.rvs()
                sizes = self.sizes.rvs(n)
                capacity = self.capacity.rvs()
                if self.fix_items:
                    self.ref_data = BinPackData(sizes, capacity)
            else:
                n = self.ref_data.sizes.shape[0]
                sizes = self.ref_data.sizes
                capacity = self.ref_data.capacity

            sizes = sizes * self.sizes_jitter.rvs(n)
            capacity = capacity * self.capacity_jitter.rvs()
            return BinPackData(sizes.round(2), capacity.round(2))

        return [_sample() for n in range(n_samples)]


def build_binpack_model(data: BinPackData) -> GurobiModel:
    """Converts bin packing problem data into a concrete Gurobipy model."""
    model = gp.Model()
    n = data.sizes.shape[0]

    # Var: Use bin
    y = model.addVars(n, name="y", vtype=GRB.BINARY)

    # Var: Assign item to bin
    x = model.addVars(n, n, name="x", vtype=GRB.BINARY)

    # Obj: Minimize number of bins
    model.setObjective(quicksum(y[i] for i in range(n)))

    # Eq: Enforce bin capacity
    model.addConstrs(
        (
            quicksum(data.sizes[i] * x[i, j] for i in range(n)) <= data.capacity * y[j]
            for j in range(n)
        ),
        name="eq_capacity",
    )

    # Eq: Must assign all items to bins
    model.addConstrs(
        (quicksum(x[i, j] for j in range(n)) == 1 for i in range(n)),
        name="eq_assign",
    )

    model.update()
    return GurobiModel(model)



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

from dataclasses import dataclass
from typing import List

import gurobipy as gp
import numpy as np
from gurobipy import GRB, quicksum
from networkx import Graph
from scipy.stats import uniform, randint
from scipy.stats.distributions import rv_frozen

from .stab import MaxWeightStableSetGenerator
from miplearn.solvers.gurobi import GurobiModel


@dataclass
class MinWeightVertexCoverData:
    graph: Graph
    weights: np.ndarray


class MinWeightVertexCoverGenerator:
    def __init__(
        self,
        w: rv_frozen = uniform(loc=10.0, scale=1.0),
        n: rv_frozen = randint(low=250, high=251),
        p: rv_frozen = uniform(loc=0.05, scale=0.0),
        fix_graph: bool = True,
    ):
        self._generator = MaxWeightStableSetGenerator(w, n, p, fix_graph)

    def generate(self, n_samples: int) -> List[MinWeightVertexCoverData]:
        return [
            MinWeightVertexCoverData(s.graph, s.weights)
            for s in self._generator.generate(n_samples)
        ]


def build_vertexcover_model(data: MinWeightVertexCoverData) -> GurobiModel:
    model = gp.Model()
    nodes = list(data.graph.nodes)
    x = model.addVars(nodes, vtype=GRB.BINARY, name="x")
    model.setObjective(quicksum(data.weights[i] * x[i] for i in nodes))
    for (v1, v2) in data.graph.edges:
        model.addConstr(x[v1] + x[v2] >= 1)
    model.update()
    return GurobiModel(model)



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.
from typing import List, Any, Callable, Optional

import numpy as np
import sklearn
from sklearn.base import BaseEstimator
from sklearn.utils.multiclass import unique_labels


class MinProbabilityClassifier(BaseEstimator):
    """
    Meta-classifier that returns NaN for predictions made by a base classifier that
    have probability below a given threshold. More specifically, this meta-classifier
    calls base_clf.predict_proba and compares the result against the provided
    thresholds. If the probability for one of the classes is above its threshold,
    the meta-classifier returns that prediction. Otherwise, it returns NaN.
    """

    def __init__(
        self,
        base_clf: Any,
        thresholds: List[float],
        clone_fn: Callable[[Any], Any] = sklearn.base.clone,
    ) -> None:
        assert len(thresholds) == 2
        self.base_clf = base_clf
        self.thresholds = thresholds
        self.clone_fn = clone_fn
        self.clf_: Optional[Any] = None
        self.classes_: Optional[List[Any]] = None

    def fit(self, x: np.ndarray, y: np.ndarray) -> None:
        assert len(y.shape) == 1
        assert len(x.shape) == 2
        classes = unique_labels(y)
        assert len(classes) == len(self.thresholds)

        self.clf_ = self.clone_fn(self.base_clf)
        self.clf_.fit(x, y)
        self.classes_ = self.clf_.classes_

    def predict(self, x: np.ndarray) -> np.ndarray:
        assert self.clf_ is not None
        assert self.classes_ is not None

        y_proba = self.clf_.predict_proba(x)
        assert len(y_proba.shape) == 2
        assert y_proba.shape[0] == x.shape[0]
        assert y_proba.shape[1] == 2
        n_samples = x.shape[0]

        y_pred = []
        for sample_idx in range(n_samples):
            yi = float("nan")
            for (class_idx, class_val) in enumerate(self.classes_):
                if y_proba[sample_idx, class_idx] >= self.thresholds[class_idx]:
                    yi = class_val
            y_pred.append(yi)
        return np.array(y_pred)



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.
import logging
from typing import List, Dict, Any
from unittest.mock import Mock

import numpy as np
from sklearn.dummy import DummyClassifier

from miplearn.components.primal.actions import SetWarmStart
from miplearn.components.primal.mem import (
    MemorizingPrimalComponent,
    SelectTopSolutions,
    MergeTopSolutions,
)
from miplearn.extractors.abstract import FeaturesExtractor

logger = logging.getLogger(__name__)


def test_mem_component(
    multiknapsack_h5: List[str], default_extractor: FeaturesExtractor
) -> None:
    # Create mock classifier
    clf = Mock(wraps=DummyClassifier())

    # Create and fit component
    comp = MemorizingPrimalComponent(
        clf,
        extractor=default_extractor,
        constructor=SelectTopSolutions(2),
        action=SetWarmStart(),
    )
    comp.fit(multiknapsack_h5)

    # Should call fit method with correct arguments
    clf.fit.assert_called()
    x, y = clf.fit.call_args.args
    assert x.shape == (3, 100)
    assert y.tolist() == [0, 1, 2]

    # Should store solutions
    assert comp.solutions_ is not None
    assert comp.solutions_.shape == (3, 100)
    assert comp.bin_var_names_ is not None
    assert len(comp.bin_var_names_) == 100

    # Call before-mip
    stats: Dict[str, Any] = {}
    model = Mock()
    comp.before_mip(multiknapsack_h5[0], model, stats)

    # Should call predict_proba with correct args
    clf.predict_proba.assert_called()
    (x_test,) = clf.predict_proba.call_args.args
    assert x_test.shape == (1, 100)

    # Should set warm starts
    model.set_warm_starts.assert_called()
    names, starts, _ = model.set_warm_starts.call_args.args
    assert len(names) == 100
    assert starts.shape == (2, 100)
    assert np.all(starts[0, :] == comp.solutions_[0, :])
    assert np.all(starts[1, :] == comp.solutions_[1, :])


def test_merge_top_solutions() -> None:
    solutions = np.array(
        [
            [0, 1, 0, 0],
            [0, 1, 0, 1],
            [0, 1, 1, 1],
            [0, 1, 1, 1],
            [1, 1, 1, 1],
        ]
    )
    y_proba = np.array([0.25, 0.25, 0.25, 0.25, 0])
    starts = MergeTopSolutions(k=4, thresholds=[0.25, 0.75]).construct(
        y_proba, solutions
    )
    assert starts.shape == (1, 4)
    assert starts[0, 0] == 0
    assert starts[0, 1] == 1
    assert np.isnan(starts[0, 2])
    assert starts[0, 3] == 1



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.
from typing import Dict, Optional, Callable, Any, List

import gurobipy as gp
from gurobipy import GRB, GurobiError
import numpy as np
from scipy.sparse import lil_matrix

from miplearn.h5 import H5File


class GurobiModel:
    _supports_basis_status = True
    _supports_sensitivity_analysis = True
    _supports_node_count = True
    _supports_solution_pool = True

    def __init__(
        self,
        inner: gp.Model,
        find_violations: Optional[Callable] = None,
        fix_violations: Optional[Callable] = None,
    ) -> None:
        self.fix_violations = fix_violations
        self.find_violations = find_violations
        self.inner = inner
        self.violations_: Optional[List[Any]] = None

    def add_constrs(
        self,
        var_names: np.ndarray,
        constrs_lhs: np.ndarray,
        constrs_sense: np.ndarray,
        constrs_rhs: np.ndarray,
        stats: Optional[Dict] = None,
    ) -> None:
        assert len(var_names.shape) == 1
        nvars = len(var_names)
        assert len(constrs_lhs.shape) == 2
        nconstrs = constrs_lhs.shape[0]
        assert constrs_lhs.shape[1] == nvars
        assert constrs_sense.shape == (nconstrs,)
        assert constrs_rhs.shape == (nconstrs,)

        gp_vars = [self.inner.getVarByName(var_name.decode()) for var_name in var_names]
        self.inner.addMConstr(constrs_lhs, gp_vars, constrs_sense, constrs_rhs)

        if stats is not None:
            if "Added constraints" not in stats:
                stats["Added constraints"] = 0
            stats["Added constraints"] += nconstrs

    def extract_after_load(self, h5: H5File) -> None:
        """
        Given a model that has just been loaded, extracts static problem
        features, such as variable names and types, objective coefficients, etc.
        """
        self.inner.update()
        self._extract_after_load_vars(h5)
        self._extract_after_load_constrs(h5)
        h5.put_scalar("static_sense", "min" if self.inner.modelSense > 0 else "max")
        h5.put_scalar("static_obj_offset", self.inner.objCon)

    def extract_after_lp(self, h5: H5File) -> None:
        """
        Given a linear programming model that has just been solved, extracts
        dynamic problem features, such as optimal LP solution, basis status,
        etc.
        """
        self._extract_after_lp_vars(h5)
        self._extract_after_lp_constrs(h5)
        h5.put_scalar("lp_obj_value", self.inner.objVal)
        h5.put_scalar("lp_wallclock_time", self.inner.runtime)

    def extract_after_mip(self, h5: H5File) -> None:
        """
        Given a mixed-integer linear programming model that has just been
        solved, extracts dynamic problem features, such as optimal MIP solution.
        """
        h5.put_scalar("mip_wallclock_time", self.inner.runtime)
        h5.put_scalar("mip_node_count", self.inner.nodeCount)
        if self.inner.status == GRB.INFEASIBLE:
            return
        gp_vars = self.inner.getVars()
        gp_constrs = self.inner.getConstrs()
        h5.put_array(
            "mip_var_values",
            np.array(self.inner.getAttr("x", gp_vars), dtype=float),
        )
        h5.put_array(
            "mip_constr_slacks",
            np.array(self.inner.getAttr("slack", gp_constrs), dtype=float),
        )
        h5.put_scalar("mip_obj_value", self.inner.objVal)
        h5.put_scalar("mip_obj_bound", self.inner.objBound)
        try:
            h5.put_scalar("mip_gap", self.inner.mipGap)
        except AttributeError:
            pass
        self._extract_after_mip_solution_pool(h5)

    def fix_variables(
        self,
        var_names: np.ndarray,
        var_values: np.ndarray,
        stats: Optional[Dict] = None,
    ) -> None:
        assert len(var_values.shape) == 1
        assert len(var_values.shape) == 1
        assert var_names.shape == var_values.shape

        n_fixed = 0
        for (var_idx, var_name) in enumerate(var_names):
            var_val = var_values[var_idx]
            if np.isfinite(var_val):
                var = self.inner.getVarByName(var_name.decode())
                var.vtype = "C"
                var.lb = var_val
                var.ub = var_val
                n_fixed += 1
        if stats is not None:
            stats["Fixed variables"] = n_fixed

    def optimize(self) -> None:
        self.violations_ = []

        def callback(m: gp.Model, where: int) -> None:
            assert self.find_violations is not None
            assert self.violations_ is not None
            assert self.fix_violations is not None
            if where == GRB.Callback.MIPSOL:
                violations = self.find_violations(self)
                self.violations_.extend(violations)
                self.fix_violations(self, violations, "cb")

        if self.fix_violations is not None:
            self.inner.Params.lazyConstraints = 1
            self.inner.optimize(callback)
        else:
            self.inner.optimize()

    def relax(self) -> "GurobiModel":
        return GurobiModel(self.inner.relax())

    def set_time_limit(self, time_limit_sec: float) -> None:
        self.inner.params.timeLimit = time_limit_sec

    def set_warm_starts(
        self,
        var_names: np.ndarray,
        var_values: np.ndarray,
        stats: Optional[Dict] = None,
    ) -> None:
        assert len(var_values.shape) == 2
        (n_starts, n_vars) = var_values.shape
        assert len(var_names.shape) == 1
        assert var_names.shape[0] == n_vars

        self.inner.numStart = n_starts
        for start_idx in range(n_starts):
            self.inner.params.startNumber = start_idx
            for (var_idx, var_name) in enumerate(var_names):
                var_val = var_values[start_idx, var_idx]
                if np.isfinite(var_val):
                    var = self.inner.getVarByName(var_name.decode())
                    var.start = var_val

        if stats is not None:
            stats["WS: Count"] = n_starts
            stats["WS: Number of variables set"] = (
                np.isfinite(var_values).mean(axis=0).sum()
            )

    def _extract_after_load_vars(self, h5: H5File) -> None:
        gp_vars = self.inner.getVars()
        for (h5_field, gp_field) in {
            "static_var_names": "varName",
            "static_var_types": "vtype",
        }.items():
            h5.put_array(
                h5_field, np.array(self.inner.getAttr(gp_field, gp_vars), dtype="S")
            )
        for (h5_field, gp_field) in {
            "static_var_upper_bounds": "ub",
            "static_var_lower_bounds": "lb",
            "static_var_obj_coeffs": "obj",
        }.items():
            h5.put_array(
                h5_field, np.array(self.inner.getAttr(gp_field, gp_vars), dtype=float)
            )

    def _extract_after_load_constrs(self, h5: H5File) -> None:
        gp_constrs = self.inner.getConstrs()
        gp_vars = self.inner.getVars()
        rhs = np.array(self.inner.getAttr("rhs", gp_constrs), dtype=float)
        senses = np.array(self.inner.getAttr("sense", gp_constrs), dtype="S")
        names = np.array(self.inner.getAttr("constrName", gp_constrs), dtype="S")
        nrows, ncols = len(gp_constrs), len(gp_vars)
        tmp = lil_matrix((nrows, ncols), dtype=float)
        for (i, gp_constr) in enumerate(gp_constrs):
            expr = self.inner.getRow(gp_constr)
            for j in range(expr.size()):
                tmp[i, expr.getVar(j).index] = expr.getCoeff(j)
        lhs = tmp.tocoo()

        h5.put_array("static_constr_names", names)
        h5.put_array("static_constr_rhs", rhs)
        h5.put_array("static_constr_sense", senses)
        h5.put_sparse("static_constr_lhs", lhs)

    def _extract_after_lp_vars(self, h5: H5File) -> None:
        def _parse_gurobi_vbasis(b: int) -> str:
            if b == 0:
                return "B"
            elif b == -1:
                return "L"
            elif b == -2:
                return "U"
            elif b == -3:
                return "S"
            else:
                raise Exception(f"unknown vbasis: {b}")

        gp_vars = self.inner.getVars()
        h5.put_array(
            "lp_var_basis_status",
            np.array(
                [
                    _parse_gurobi_vbasis(b)
                    for b in self.inner.getAttr("vbasis", gp_vars)
                ],
                dtype="S",
            ),
        )
        for (h5_field, gp_field) in {
            "lp_var_reduced_costs": "rc",
            "lp_var_sa_obj_up": "saobjUp",
            "lp_var_sa_obj_down": "saobjLow",
            "lp_var_sa_ub_up": "saubUp",
            "lp_var_sa_ub_down": "saubLow",
            "lp_var_sa_lb_up": "salbUp",
            "lp_var_sa_lb_down": "salbLow",
            "lp_var_values": "x",
        }.items():
            h5.put_array(
                h5_field,
                np.array(self.inner.getAttr(gp_field, gp_vars), dtype=float),
            )

    def _extract_after_lp_constrs(self, h5: H5File) -> None:
        def _parse_gurobi_cbasis(v: int) -> str:
            if v == 0:
                return "B"
            if v == -1:
                return "N"
            raise Exception(f"unknown cbasis: {v}")

        gp_constrs = self.inner.getConstrs()
        h5.put_array(
            "lp_constr_basis_status",
            np.array(
                [
                    _parse_gurobi_cbasis(c)
                    for c in self.inner.getAttr("cbasis", gp_constrs)
                ],
                dtype="S",
            ),
        )
        for (h5_field, gp_field) in {
            "lp_constr_dual_values": "pi",
            "lp_constr_sa_rhs_up": "saRhsUp",
            "lp_constr_sa_rhs_down": "saRhsLow",
            "lp_constr_slacks": "slack",
        }.items():
            h5.put_array(
                h5_field,
                np.array(self.inner.getAttr(gp_field, gp_constrs), dtype=float),
            )

    def _extract_after_mip_solution_pool(self, h5: H5File) -> None:
        gp_vars = self.inner.getVars()
        pool_var_values = []
        pool_obj_values = []
        for i in range(self.inner.SolCount):
            self.inner.params.SolutionNumber = i
            try:
                pool_var_values.append(self.inner.getAttr("Xn", gp_vars))
                pool_obj_values.append(self.inner.PoolObjVal)
            except GurobiError:
                pass
        h5.put_array("pool_var_values", np.array(pool_var_values))
        h5.put_array("pool_obj_values", np.array(pool_obj_values))

    def write(self, filename: str) -> None:
        self.inner.update()
        self.inner.write(filename)



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

from dataclasses import dataclass
from typing import List, Optional

import gurobipy as gp
import numpy as np
from gurobipy import GRB
from scipy.stats import uniform, randint
from scipy.stats.distributions import rv_frozen

from miplearn.solvers.gurobi import GurobiModel


@dataclass
class MultiKnapsackData:
    """Data for the multi-dimensional knapsack problem

    Args
    ----
    prices
        Item prices.
    capacities
        Knapsack capacities.
    weights
        Matrix of item weights.
    """

    prices: np.ndarray
    capacities: np.ndarray
    weights: np.ndarray


# noinspection PyPep8Naming
class MultiKnapsackGenerator:
    """Random instance generator for the multi-dimensional knapsack problem.

    Instances have a random number of items (or variables) and a random number of
    knapsacks (or constraints), as specified by the provided probability
    distributions `n` and `m`, respectively. The weight of each item `i` on knapsack
    `j` is sampled independently from the provided distribution `w`. The capacity of
    knapsack `j` is set to ``alpha_j * sum(w[i,j] for i in range(n))``,
    where `alpha_j`, the tightness ratio, is sampled from the provided probability
    distribution `alpha`.

    To make the instances more challenging, the costs of the items are linearly
    correlated to their average weights. More specifically, the weight of each item
    `i` is set to ``sum(w[i,j]/m for j in range(m)) + K * u_i``, where `K`,
    the correlation coefficient, and `u_i`, the correlation multiplier, are sampled
    from the provided probability distributions. Note that `K` is only sample once
    for the entire instance.

    If `fix_w=True`, then `weights[i,j]` are kept the same in all generated
    instances. This also implies that n and m are kept fixed. Although the prices and
    capacities are derived from `weights[i,j]`, as long as `u` and `K` are not
    constants, the generated instances will still not be completely identical.

    If a probability distribution `w_jitter` is provided, then item weights will be
    set to ``w[i,j] * gamma[i,j]`` where `gamma[i,j]` is sampled from `w_jitter`.
    When combined with `fix_w=True`, this argument may be used to generate instances
    where the weight of each item is roughly the same, but not exactly identical,
    across all instances. The prices of the items and the capacities of the knapsacks
    will be calculated as above, but using these perturbed weights instead.

    By default, all generated prices, weights and capacities are rounded to the
    nearest integer number. If `round=False` is provided, this rounding will be
    disabled.

    Parameters
    ----------
    n: rv_discrete
        Probability distribution for the number of items (or variables).
    m: rv_discrete
        Probability distribution for the number of knapsacks (or constraints).
    w: rv_continuous
        Probability distribution for the item weights.
    K: rv_continuous
        Probability distribution for the profit correlation coefficient.
    u: rv_continuous
        Probability distribution for the profit multiplier.
    alpha: rv_continuous
        Probability distribution for the tightness ratio.
    fix_w: boolean
        If true, weights are kept the same (minus the noise from w_jitter) in all
        instances.
    w_jitter: rv_continuous
        Probability distribution for random noise added to the weights.
    round: boolean
        If true, all prices, weights and capacities are rounded to the nearest
        integer.
    """

    def __init__(
        self,
        n: rv_frozen = randint(low=100, high=101),
        m: rv_frozen = randint(low=30, high=31),
        w: rv_frozen = randint(low=0, high=1000),
        K: rv_frozen = randint(low=500, high=501),
        u: rv_frozen = uniform(loc=0.0, scale=1.0),
        alpha: rv_frozen = uniform(loc=0.25, scale=0.0),
        fix_w: bool = False,
        w_jitter: rv_frozen = uniform(loc=1.0, scale=0.0),
        p_jitter: rv_frozen = uniform(loc=1.0, scale=0.0),
        round: bool = True,
    ):
        assert isinstance(n, rv_frozen), "n should be a SciPy probability distribution"
        assert isinstance(m, rv_frozen), "m should be a SciPy probability distribution"
        assert isinstance(w, rv_frozen), "w should be a SciPy probability distribution"
        assert isinstance(K, rv_frozen), "K should be a SciPy probability distribution"
        assert isinstance(u, rv_frozen), "u should be a SciPy probability distribution"
        assert isinstance(
            alpha, rv_frozen
        ), "alpha should be a SciPy probability distribution"
        assert isinstance(fix_w, bool), "fix_w should be boolean"
        assert isinstance(
            w_jitter, rv_frozen
        ), "w_jitter should be a SciPy probability distribution"

        self.n = n
        self.m = m
        self.w = w
        self.u = u
        self.K = K
        self.alpha = alpha
        self.w_jitter = w_jitter
        self.p_jitter = p_jitter
        self.round = round
        self.fix_n: Optional[int] = None
        self.fix_m: Optional[int] = None
        self.fix_w: Optional[np.ndarray] = None
        self.fix_u: Optional[np.ndarray] = None
        self.fix_K: Optional[float] = None

        if fix_w:
            self.fix_n = self.n.rvs()
            self.fix_m = self.m.rvs()
            self.fix_w = np.array([self.w.rvs(self.fix_n) for _ in range(self.fix_m)])
            self.fix_u = self.u.rvs(self.fix_n)
            self.fix_K = self.K.rvs()

    def generate(self, n_samples: int) -> List[MultiKnapsackData]:
        def _sample() -> MultiKnapsackData:
            if self.fix_w is not None:
                assert self.fix_m is not None
                assert self.fix_n is not None
                assert self.fix_u is not None
                assert self.fix_K is not None
                n = self.fix_n
                m = self.fix_m
                w = self.fix_w
                u = self.fix_u
                K = self.fix_K
            else:
                n = self.n.rvs()
                m = self.m.rvs()
                w = np.array([self.w.rvs(n) for _ in range(m)])
                u = self.u.rvs(n)
                K = self.K.rvs()
            w = w * np.array([self.w_jitter.rvs(n) for _ in range(m)])
            alpha = self.alpha.rvs(m)
            p = np.array(
                [w[:, j].sum() / m + K * u[j] for j in range(n)]
            ) * self.p_jitter.rvs(n)
            b = np.array([w[i, :].sum() * alpha[i] for i in range(m)])
            if self.round:
                p = p.round()
                b = b.round()
                w = w.round()
            return MultiKnapsackData(p, b, w)

        return [_sample() for _ in range(n_samples)]


def build_multiknapsack_model(data: MultiKnapsackData) -> GurobiModel:
    """Converts multi-knapsack problem data into a concrete Gurobipy model."""
    model = gp.Model()
    m, n = data.weights.shape
    x = model.addMVar(n, vtype=GRB.BINARY, name="x")
    model.addConstr(data.weights @ x <= data.capacities)
    model.setObjective(-data.prices @ x)
    model.update()
    return GurobiModel(model)



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.
from typing import Callable, Optional

import numpy as np
import sklearn.base
from sklearn.base import BaseEstimator
from sklearn.utils.multiclass import unique_labels


class SingleClassFix(BaseEstimator):
    """
    Some sklearn classifiers, such as logistic regression, have issues with datasets
    that contain a single class. This meta-classifier fixes the issue. If the
    training data contains a single class, this meta-classifier always returns that
    class as a prediction. Otherwise, it fits the provided base classifier,
    and returns its predictions instead.
    """

    def __init__(
        self,
        base_clf: BaseEstimator,
        clone_fn: Callable = sklearn.base.clone,
    ):
        self.base_clf = base_clf
        self.clf_: Optional[BaseEstimator] = None
        self.constant_ = None
        self.classes_ = None
        self.clone_fn = clone_fn

    def fit(self, x: np.ndarray, y: np.ndarray) -> None:
        classes = unique_labels(y)
        if len(classes) == 1:
            assert classes[0] is not None
            self.clf_ = None
            self.constant_ = classes[0]
            self.classes_ = classes
        else:
            self.clf_ = self.clone_fn(self.base_clf)
            assert self.clf_ is not None
            self.clf_.fit(x, y)
            self.constant_ = None
            self.classes_ = self.clf_.classes_

    def predict(self, x: np.ndarray) -> np.ndarray:
        if self.constant_ is not None:
            return np.full(x.shape[0], self.constant_)
        else:
            assert self.clf_ is not None
            return self.clf_.predict(x)



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

import json
from typing import Any, Dict, List

import gurobipy as gp

from ..h5 import H5File


class ExpertLazyComponent:
    def __init__(self) -> None:
        pass

    def fit(self, train_h5: List[str]) -> None:
        pass

    def before_mip(self, test_h5: str, model: gp.Model, stats: Dict[str, Any]) -> None:
        with H5File(test_h5, "r") as h5:
            constr_names = h5.get_array("static_constr_names")
            constr_lazy = h5.get_array("mip_constr_lazy")
            constr_violations = h5.get_scalar("mip_constr_violations")

            assert constr_names is not None
            assert constr_violations is not None

            # Static lazy constraints
            n_static_lazy = 0
            if constr_lazy is not None:
                for (constr_idx, constr_name) in enumerate(constr_names):
                    if constr_lazy[constr_idx]:
                        constr = model.getConstrByName(constr_name.decode())
                        constr.lazy = 3
                        n_static_lazy += 1
            stats.update({"Static lazy constraints": n_static_lazy})

            # Dynamic lazy constraints
            if hasattr(model, "_fix_violations"):
                violations = json.loads(constr_violations)
                model._fix_violations(model, violations, "aot")
                stats.update({"Dynamic lazy constraints": len(violations)})



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

from tempfile import NamedTemporaryFile

import numpy as np
import pytest

from miplearn.h5 import H5File
from miplearn.problems.setcover import (
    SetCoverData,
    build_setcover_model_gurobipy,
    build_setcover_model_pyomo,
)
from miplearn.solvers.abstract import AbstractModel

inf = float("inf")


@pytest.fixture
def data() -> SetCoverData:
    return SetCoverData(
        costs=np.array([5, 10, 12, 6, 8]),
        incidence_matrix=np.array(
            [
                [1, 0, 0, 1, 0],
                [1, 1, 0, 0, 0],
                [0, 0, 1, 1, 1],
            ],
        ),
    )


def test_gurobi(data: SetCoverData) -> None:
    _test_solver(build_setcover_model_gurobipy, data)


def test_pyomo_persistent(data: SetCoverData) -> None:
    _test_solver(lambda d: build_setcover_model_pyomo(d, "gurobi_persistent"), data)


def _test_solver(build_model, data):
    _test_extract(build_model(data))
    _test_add_constr(build_model(data))
    _test_fix_vars(build_model(data))
    _test_infeasible(build_model(data))


def _test_extract(model):
    with NamedTemporaryFile() as tempfile:
        with H5File(tempfile.name) as h5:

            def test_scalar(key, expected_value):
                actual_value = h5.get_scalar(key)
                assert actual_value is not None
                assert actual_value == expected_value

            def test_array(key, expected_value):
                actual_value = h5.get_array(key)
                assert actual_value is not None
                assert actual_value.tolist() == expected_value

            def test_sparse(key, expected_value):
                actual_value = h5.get_sparse(key)
                assert actual_value is not None
                assert actual_value.todense().tolist() == expected_value

            model.extract_after_load(h5)
            test_sparse(
                "static_constr_lhs",
                [
                    [1.0, 0.0, 0.0, 1.0, 0.0],
                    [1.0, 1.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0],
                ],
            )
            test_array("static_constr_names", [b"eqs[0]", b"eqs[1]", b"eqs[2]"])
            test_array("static_constr_rhs", [1, 1, 1])
            test_array("static_constr_sense", [b">", b">", b">"])
            test_scalar("static_obj_offset", 0.0)
            test_scalar("static_sense", "min")
            test_array("static_var_lower_bounds", [0.0, 0.0, 0.0, 0.0, 0.0])
            test_array(
                "static_var_names",
                [
                    b"x[0]",
                    b"x[1]",
                    b"x[2]",
                    b"x[3]",
                    b"x[4]",
                ],
            )
            test_array("static_var_obj_coeffs", [5.0, 10.0, 12.0, 6.0, 8.0])
            test_array("static_var_types", [b"B", b"B", b"B", b"B", b"B"])
            test_array("static_var_upper_bounds", [1.0, 1.0, 1.0, 1.0, 1.0])

            relaxed = model.relax()
            relaxed.optimize()
            relaxed.extract_after_lp(h5)
            test_array("lp_constr_dual_values", [0, 5, 6])
            test_array("lp_constr_slacks", [-1, 0, 0])
            test_scalar("lp_obj_value", 11.0)
            test_array("lp_var_reduced_costs", [0.0, 5.0, 6.0, 0.0, 2.0])
            test_array("lp_var_values", [1.0, 0.0, 0.0, 1.0, 0.0])
            if model._supports_basis_status:
                test_array("lp_var_basis_status", [b"B", b"L", b"L", b"B", b"L"])
                test_array("lp_constr_basis_status", [b"B", b"N", b"N"])
            if model._supports_sensitivity_analysis:
                test_array("lp_constr_sa_rhs_up", [2, 1, 1])
                test_array("lp_constr_sa_rhs_down", [-inf, 0, 0])
                test_array("lp_var_sa_obj_up", [10.0, inf, inf, 8.0, inf])
                test_array("lp_var_sa_obj_down", [0.0, 5.0, 6.0, 0.0, 6.0])
                test_array("lp_var_sa_ub_up", [inf, inf, inf, inf, inf])
                test_array("lp_var_sa_ub_down", [1.0, 0.0, 0.0, 1.0, 0.0])
                test_array("lp_var_sa_lb_up", [1.0, 1.0, 1.0, 1.0, 1.0])
                test_array("lp_var_sa_lb_down", [-inf, 0.0, 0.0, -inf, 0.0])
            lp_wallclock_time = h5.get_scalar("lp_wallclock_time")
            assert lp_wallclock_time is not None
            assert lp_wallclock_time >= 0

            model.optimize()
            model.extract_after_mip(h5)
            test_array("mip_constr_slacks", [-1, 0, 0])
            test_array("mip_var_values", [1.0, 0.0, 0.0, 1.0, 0.0])
            test_scalar("mip_gap", 0)
            test_scalar("mip_obj_bound", 11.0)
            test_scalar("mip_obj_value", 11.0)
            mip_wallclock_time = h5.get_scalar("mip_wallclock_time")
            assert mip_wallclock_time is not None
            assert mip_wallclock_time > 0
            if model._supports_node_count:
                count = h5.get_scalar("mip_node_count")
                assert count is not None
                assert count >= 0
            if model._supports_solution_pool:
                pool_var_values = h5.get_array("pool_var_values")
                pool_obj_values = h5.get_array("pool_obj_values")
                assert pool_var_values is not None
                assert pool_obj_values is not None
                assert len(pool_obj_values.shape) == 1
                n_sols = len(pool_obj_values)
                assert pool_var_values.shape == (n_sols, 5)


def _test_add_constr(model: AbstractModel):
    with NamedTemporaryFile() as tempfile:
        with H5File(tempfile.name) as h5:
            model.add_constrs(
                var_names=np.array([b"x[2]", b"x[3]"], dtype="S"),
                constrs_lhs=np.array([[0, 1], [1, 0]]),
                constrs_sense=np.array(["=", "="], dtype="S"),
                constrs_rhs=np.array([0, 0]),
            )
            model.optimize()
            model.extract_after_mip(h5)
            assert h5.get_array("mip_var_values").tolist() == [1, 0, 0, 0, 1]


def _test_fix_vars(model: AbstractModel):
    with NamedTemporaryFile() as tempfile:
        with H5File(tempfile.name) as h5:
            model.fix_variables(
                var_names=np.array([b"x[2]", b"x[3]"], dtype="S"),
                var_values=np.array([0, 0]),
            )
            model.optimize()
            model.extract_after_mip(h5)
            assert h5.get_array("mip_var_values").tolist() == [1, 0, 0, 0, 1]


def _test_infeasible(model: AbstractModel):
    with NamedTemporaryFile() as tempfile:
        with H5File(tempfile.name) as h5:
            model.fix_variables(
                var_names=np.array([b"x[0]", b"x[3]"], dtype="S"),
                var_values=np.array([0, 0]),
            )
            model.optimize()
            model.extract_after_mip(h5)
            assert h5.get_array("mip_var_values") is None



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

import random
from dataclasses import dataclass
from typing import Any, Callable, List

import numpy as np
from scipy.stats import randint, uniform
from sklearn.decomposition import PCA
from sklearn.dummy import DummyClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.neighbors import KNeighborsClassifier
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler

from sklearn.tree import DecisionTreeClassifier

from miplearn.classifiers.minprob import MinProbabilityClassifier
from miplearn.classifiers.singleclass import SingleClassFix
from miplearn.collectors.basic import BasicCollector
from miplearn.components.primal.actions import (
    SetWarmStart,
    FixVariables,
    EnforceProximity,
)
from miplearn.components.primal.expert import ExpertPrimalComponent
from miplearn.components.primal.indep import IndependentVarsPrimalComponent
from miplearn.components.primal.joint import JointVarsPrimalComponent
from miplearn.components.primal.mem import (
    MemorizingPrimalComponent,
    SelectTopSolutions,
    MergeTopSolutions,
)
from miplearn.extractors.fields import H5FieldsExtractor
from miplearn.problems.binpack import BinPackGenerator, build_binpack_model
from miplearn.problems.multiknapsack import (
    MultiKnapsackGenerator,
    build_multiknapsack_model,
)
from miplearn.problems.pmedian import PMedianGenerator, build_pmedian_model
from miplearn.problems.setcover import SetCoverGenerator, build_setcover_model_gurobipy
from miplearn.problems.setpack import SetPackGenerator, build_setpack_model
from miplearn.problems.stab import (
    MaxWeightStableSetGenerator,
    build_stab_model_gurobipy,
    build_stab_model_pyomo,
)
from miplearn.problems.tsp import TravelingSalesmanGenerator, build_tsp_model
from miplearn.problems.uc import UnitCommitmentGenerator, build_uc_model
from miplearn.problems.vertexcover import (
    MinWeightVertexCoverGenerator,
    build_vertexcover_model,
)

dev = True

n_jobs = 8 if dev else 32
timelimit_collect = 900
timelimit_benchmark = 300
n_seeds = 1 if dev else 5

random.seed(42)
np.random.seed(42)


@dataclass
class Challenge:
    name: str
    generator: Any
    build_model: Callable
    n_samples: int = 16 if dev else 1056
    n_train: int = 0 if dev else 1024


@dataclass
class Method:
    name: str
    components: List


# CHALLENGES
# -----------------------------------------------------------------------------

challenges: List[Challenge] = []

# # Bin Packing
# for i in [] if dev else range(5, 10):
#     n = i * 100
#     challenges.append(
#         Challenge(
#             name=f"binpack-n{n:03d}",
#             generator=BinPackGenerator(
#                 n=randint(low=n, high=n + 1),
#                 sizes=uniform(loc=0, scale=25),
#                 capacity=uniform(loc=100, scale=0),
#                 sizes_jitter=uniform(loc=0.9, scale=0.2),
#                 capacity_jitter=uniform(loc=0.9, scale=0.2),
#                 fix_items=True,
#             ),
#             build_model=build_binpack_model,
#         )
#     )
#
# # Multi-knapsack
# for i in [3] if dev else range(5, 10):
#     n = 25 * i
#     m = i
#     challenges.append(
#         Challenge(
#             name=f"multiknapsack-n{n:03d}-m{m}",
#             generator=MultiKnapsackGenerator(
#                 n=randint(low=n, high=n + 1),
#                 m=randint(low=m, high=m + 1),
#                 w=uniform(loc=0, scale=1000),
#                 K=uniform(loc=100, scale=0),
#                 u=uniform(loc=1, scale=0),
#                 alpha=uniform(loc=0.25, scale=0),
#                 w_jitter=uniform(loc=0.95, scale=0.1),
#                 p_jitter=uniform(loc=0.75, scale=0.5),
#                 fix_w=True,
#             ),
#             build_model=build_multiknapsack_model,
#         )
#     )
#
# # Capacitated p-median
# for i in [] if dev else range(5, 10):
#     n = 50 * i
#     p = i
#     challenges.append(
#         Challenge(
#             name=f"pmedian-n{n:03d}-p{p:01d}",
#             generator=PMedianGenerator(
#                 x=uniform(loc=0.0, scale=100.0),
#                 y=uniform(loc=0.0, scale=100.0),
#                 n=randint(low=n, high=n + 1),
#                 p=randint(low=p, high=p + 1),
#                 demands=uniform(loc=0, scale=10),
#                 capacities=uniform(loc=0, scale=250),
#                 distances_jitter=uniform(loc=0.9, scale=0.2),
#                 demands_jitter=uniform(loc=0.9, scale=0.2),
#                 capacities_jitter=uniform(loc=0.9, scale=0.2),
#                 fixed=True,
#             ),
#             build_model=build_pmedian_model,
#         )
#     )
#
# # Set Cover
# for i in [7] if dev else range(9, 14):
#     n = 250 * i
#     m = 10 * i
#     challenges.append(
#         Challenge(
#             name=f"setcover-n{n:04d}-m{m:03d}",
#             generator=SetCoverGenerator(
#                 n_elements=randint(low=m, high=m + 1),
#                 n_sets=randint(low=n, high=n + 1),
#                 costs=uniform(loc=0.0, scale=1000.0),
#                 costs_jitter=uniform(loc=0.90, scale=0.20),
#                 density=uniform(loc=0.05, scale=0.00),
#                 K=uniform(loc=25.0, scale=0.0),
#                 fix_sets=True,
#             ),
#             build_model=build_setcover_model_gurobipy,
#         )
#     )
#
# # Set Pack
# for i in [12] if dev else range(14, 19):
#     n = 100 * i
#     m = 10 * i
#     challenges.append(
#         Challenge(
#             name=f"setpack-n{n:04d}-m{m:03d}",
#             generator=SetPackGenerator(
#                 n_elements=randint(low=m, high=m + 1),
#                 n_sets=randint(low=n, high=n + 1),
#                 costs=uniform(loc=0.0, scale=1000.0),
#                 costs_jitter=uniform(loc=0.75, scale=0.5),
#                 density=uniform(loc=0.10, scale=0.00),
#                 K=uniform(loc=10.0, scale=0.0),
#                 fix_sets=True,
#             ),
#             build_model=build_setpack_model,
#         )
#     )

# Maximum Weight Stable Set
for i in [16] if dev else range(18, 23):
    n = 10 * i
    challenges.append(
        Challenge(
            name=f"stab-n{n:03d}",
            generator=MaxWeightStableSetGenerator(
                w=uniform(loc=100.0, scale=25.0),
                n=randint(low=n, high=n + 1),
                p=uniform(loc=0.05, scale=0.0),
                fix_graph=True,
            ),
            build_model=build_stab_model_pyomo,
        )
    )

# # Traveling Salesman
# for i in [] if dev else range(7, 12):
#     n = i * 50
#     challenges.append(
#         Challenge(
#             name=f"tsp-n{n:03d}",
#             generator=TravelingSalesmanGenerator(
#                 n=randint(low=n, high=n + 1),
#                 x=uniform(loc=0.0, scale=1000.0),
#                 y=uniform(loc=0.0, scale=1000.0),
#                 gamma=uniform(loc=0.90, scale=0.20),
#                 fix_cities=True,
#                 round=True,
#             ),
#             build_model=build_tsp_model,
#         )
#     )

# # Unit Commitment
# for i in [] if dev else range(6, 11):
#     g = 250 * i
#     t = i * 4
#     challenges.append(
#         Challenge(
#             name=f"uc-g{g:04d}-t{t:02d}",
#             generator=UnitCommitmentGenerator(
#                 n_units=randint(low=g, high=g + 1),
#                 n_periods=randint(low=t, high=t + 1),
#                 max_power=uniform(loc=50, scale=450),
#                 min_power=uniform(loc=0.5, scale=0.25),
#                 cost_startup=uniform(loc=0, scale=10_000),
#                 cost_prod=uniform(loc=0, scale=50),
#                 cost_fixed=uniform(loc=0, scale=1_000),
#                 min_uptime=randint(low=2, high=8),
#                 min_downtime=randint(low=2, high=8),
#                 cost_jitter=uniform(loc=0.75, scale=0.5),
#                 demand_jitter=uniform(loc=0.9, scale=0.2),
#                 fix_units=True,
#             ),
#             build_model=build_uc_model,
#         )
#     )

# # Vertex Cover
# for i in [] if dev else range(16, 21):
#     n = 10 * i
#     challenges.append(
#         Challenge(
#             name=f"vertexcover-n{n:03d}",
#             generator=MinWeightVertexCoverGenerator(
#                 w=uniform(loc=100.0, scale=25.0),
#                 n=randint(low=n, high=n + 1),
#                 p=uniform(loc=0.25, scale=0),
#                 fix_graph=True,
#             ),
#             build_model=build_vertexcover_model,
#         )
#     )

# COLLECTORS
# -----------------------------------------------------------------------------
collectors = [
    BasicCollector(),
    # BranchPriorityCollector(
    #     time_limit=timelimit_collect,
    #     node_limit=500,
    # ),
    # LazyCollector(
    #     time_limit=timelimit_collect,
    #     min_constrs=100_000,
    # ),
]

# METHODS
# -----------------------------------------------------------------------------

# baseline & experts
methods: List[Method] = [
    Method(
        name="baseline",
        components=[],
    ),
    Method(
        name="ws:exp",
        components=[
            ExpertPrimalComponent(action=SetWarmStart()),
        ],
    ),
    Method(
        name="fix:exp",
        components=[
            ExpertPrimalComponent(action=FixVariables()),
        ],
    ),
    Method(
        name="prox:exp",
        components=[
            ExpertPrimalComponent(action=EnforceProximity(0.01)),
        ],
    ),
]

# extractor = H5FieldsExtractor(
#     instance_fields=[
#         "lp_obj_value",
#         "lp_var_values",
#         "lp_var_reduced_costs",
#         "lp_constr_dual_values",
#         "static_var_obj_coeffs",
#         "static_constr_rhs",
#     ],
#     var_fields=["lp_var_features"],
# )
#
# for k in [25]:
#     methods.append(
#         Method(
#             name=f"ws:collect:{k}",
#             components=[
#                 MemorizingPrimalComponent(
#                     clf=DecisionTreeClassifier(),
#                     constructor=SelectTopSolutions(k),
#                     extractor=extractor,
#                     action=SetWarmStart(),
#                 ),
#             ],
#         )
#     )
#     methods.append(
#         Method(
#             name=f"ws:merge:{k}",
#             components=[
#                 MemorizingPrimalComponent(
#                     clf=DecisionTreeClassifier(),
#                     constructor=MergeTopSolutions(k, thresholds=[0.05, 0.95]),
#                     extractor=extractor,
#                     action=SetWarmStart(),
#                 ),
#             ],
#         )
#     )
#     methods.append(
#         Method(
#             name=f"fix:merge:{k}",
#             components=[
#                 MemorizingPrimalComponent(
#                     clf=DecisionTreeClassifier(),
#                     constructor=MergeTopSolutions(k, thresholds=[0.05, 0.95]),
#                     extractor=extractor,
#                     action=FixVariables(),
#                 ),
#             ],
#         )
#     )
#
# for k1 in [25]:
#     for k2 in [1, 5, 10]:
#         methods.append(
#             Method(
#                 name=f"prox:merge:{k1}:{k2}",
#                 components=[
#                     MemorizingPrimalComponent(
#                         clf=DecisionTreeClassifier(),
#                         constructor=MergeTopSolutions(k1, thresholds=[0.05, 0.95]),
#                         extractor=extractor,
#                         action=EnforceProximity(k2 / 100),
#                     ),
#                 ],
#             )
#         )

# methods.append(
#     Method(
#         name=f"ws:j:tree",
#         components=[
#             JointVarsPrimalComponent(
#                 clf=DecisionTreeClassifier(),
#                 extractor=extractor,
#             ),
#         ],
#     )
# )
#
# methods.append(
#     Method(
#         name=f"ws:i:tree",
#         components=[
#             IndependentVarsPrimalComponent(
#                 base_clf=SingleClassFix(
#                     MinProbabilityClassifier(
#                         LogisticRegression(),
#                         thresholds=[0.99, 0.99],
#                     )
#                 ),
#                 extractor=extractor,
#             ),
#         ],
#     )
# )



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

import numpy as np
from scipy.stats import uniform, randint

from miplearn.problems.binpack import build_binpack_model, BinPackData, BinPackGenerator


def test_binpack_generator() -> None:
    np.random.seed(42)
    gen = BinPackGenerator(
        n=randint(low=10, high=11),
        sizes=uniform(loc=0, scale=10),
        capacity=uniform(loc=100, scale=0),
        sizes_jitter=uniform(loc=0.9, scale=0.2),
        capacity_jitter=uniform(loc=0.9, scale=0.2),
        fix_items=True,
    )
    data = gen.generate(2)
    assert data[0].sizes.tolist() == [
        3.39,
        10.4,
        7.81,
        5.64,
        1.46,
        1.46,
        0.56,
        8.7,
        5.93,
        6.79,
    ]
    assert data[0].capacity == 102.24
    assert data[1].sizes.tolist() == [
        3.48,
        9.11,
        7.12,
        5.93,
        1.65,
        1.47,
        0.58,
        8.82,
        5.47,
        7.23,
    ]
    assert data[1].capacity == 93.41


def test_binpack() -> None:
    model = build_binpack_model(
        BinPackData(
            sizes=np.array([4, 8, 1, 4, 2, 1]),
            capacity=10,
        )
    )
    model.optimize()
    assert model.inner.objVal == 2.0



from abc import ABC, abstractmethod

import numpy as np

from miplearn.h5 import H5File


class FeaturesExtractor(ABC):
    @abstractmethod
    def get_instance_features(self, h5: H5File) -> np.ndarray:
        pass

    @abstractmethod
    def get_var_features(self, h5: H5File) -> np.ndarray:
        pass

    @abstractmethod
    def get_constr_features(self, h5: H5File) -> np.ndarray:
        pass



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

from gzip import GzipFile
import os
import pickle
import sys
from typing import IO, Any, Callable, List, cast, TextIO

from .parallel import p_umap
import shutil


class _RedirectOutput:
    def __init__(self, streams: List[Any]) -> None:
        self.streams = streams

    def write(self, data: Any) -> None:
        for stream in self.streams:
            stream.write(data)

    def flush(self) -> None:
        for stream in self.streams:
            stream.flush()

    def __enter__(self) -> Any:
        self._original_stdout = sys.stdout
        self._original_stderr = sys.stderr
        sys.stdout = cast(TextIO, self)
        sys.stderr = cast(TextIO, self)
        return self

    def __exit__(
        self,
        _type: Any,
        _value: Any,
        _traceback: Any,
    ) -> None:
        sys.stdout = self._original_stdout
        sys.stderr = self._original_stderr


def save(
    objs: List[Any],
    dirname: str,
    prefix: str = "",
    n_jobs: int = 1,
    progress: bool = False,
) -> List[str]:
    filenames = [f"{dirname}/{prefix}{i:05d}.pkl.gz" for i in range(len(objs))]

    def _process(i: int) -> None:
        filename = filenames[i]
        obj = objs[i]
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        with GzipFile(filename, "wb") as file:
            pickle.dump(obj, cast(IO[bytes], file))

    p_umap(
        _process,
        range(len(objs)),
        smoothing=0,
        num_cpus=n_jobs,
        maxtasksperchild=None,
        disable=not progress,
    )
    return filenames


def load(filename: str, build_model: Callable) -> Any:
    with GzipFile(filename, "rb") as file:
        data = pickle.load(cast(IO[bytes], file))
        return build_model(data)


def gzip(filename: str) -> None:
    with open(filename, "rb") as input_file:
        with GzipFile(f"{filename}.gz", "wb") as output_file:
            shutil.copyfileobj(input_file, output_file)
    os.remove(filename)



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.
from numbers import Number
from typing import Optional, Dict, List, Any

import numpy as np
import pyomo
from pyomo.core import Objective, Var, Suffix
from pyomo.core.base import _GeneralVarData
from pyomo.core.expr.numeric_expr import SumExpression, MonomialTermExpression
from scipy.sparse import coo_matrix

from miplearn.h5 import H5File
from miplearn.solvers.abstract import AbstractModel
import pyomo.environ as pe


class PyomoModel(AbstractModel):
    def __init__(self, model: pe.ConcreteModel, solver_name: str):
        self.model = model
        self.solver_name = solver_name
        self.solver = pe.SolverFactory(solver_name)
        self.is_persistent = hasattr(self.solver, "set_instance")
        if self.is_persistent:
            self.solver.set_instance(model)
        self.results = None
        self._is_warm_start_available = False
        if not hasattr(self.model, "dual"):
            self.model.dual = Suffix(direction=Suffix.IMPORT)
            self.model.rc = Suffix(direction=Suffix.IMPORT)
            self.model.slack = Suffix(direction=Suffix.IMPORT)

    def add_constrs(
        self,
        var_names: np.ndarray,
        constrs_lhs: np.ndarray,
        constrs_sense: np.ndarray,
        constrs_rhs: np.ndarray,
        stats: Optional[Dict] = None,
    ) -> None:
        variables = self._var_names_to_vars(var_names)
        if not hasattr(self.model, "added_eqs"):
            self.model.added_eqs = pe.ConstraintList()
        for i in range(len(constrs_sense)):
            lhs = sum([variables[j] * constrs_lhs[i, j] for j in range(len(variables))])
            sense = constrs_sense[i]
            rhs = constrs_rhs[i]
            if sense == b"=":
                eq = self.model.added_eqs.add(lhs == rhs)
            elif sense == b"<":
                eq = self.model.added_eqs.add(lhs <= rhs)
            elif sense == b">":
                eq = self.model.added_eqs.add(lhs >= rhs)
            else:
                raise Exception(f"Unknown sense: {sense}")
            self.solver.add_constraint(eq)

    def _var_names_to_vars(self, var_names):
        varname_to_var = {}
        for var in self.model.component_objects(Var):
            for idx in var:
                v = var[idx]
                varname_to_var[v.name] = var[idx]
        return [varname_to_var[var_name.decode()] for var_name in var_names]

    def extract_after_load(self, h5: H5File) -> None:
        self._extract_after_load_vars(h5)
        self._extract_after_load_constrs(h5)
        h5.put_scalar("static_sense", self._get_sense())

    def extract_after_lp(self, h5: H5File) -> None:
        self._extract_after_lp_vars(h5)
        self._extract_after_lp_constrs(h5)
        h5.put_scalar("lp_obj_value", self.results["Problem"][0]["Lower bound"])
        h5.put_scalar("lp_wallclock_time", self._get_runtime())

    def _get_runtime(self):
        solver_dict = self.results["Solver"][0]
        for key in ["Wallclock time", "User time"]:
            if isinstance(solver_dict[key], Number):
                return solver_dict[key]
        raise Exception("Time unavailable")

    def extract_after_mip(self, h5: H5File) -> None:
        h5.put_scalar("mip_wallclock_time", self._get_runtime())
        if self.results["Solver"][0]["Termination condition"] == "infeasible":
            return
        self._extract_after_mip_vars(h5)
        self._extract_after_mip_constrs(h5)
        if self._get_sense() == "max":
            obj_value = self.results["Problem"][0]["Lower bound"]
            obj_bound = self.results["Problem"][0]["Upper bound"]
        else:
            obj_value = self.results["Problem"][0]["Upper bound"]
            obj_bound = self.results["Problem"][0]["Lower bound"]
        h5.put_scalar("mip_obj_value", obj_value)
        h5.put_scalar("mip_obj_bound", obj_bound)
        h5.put_scalar("mip_gap", self._gap(obj_value, obj_bound))

    def fix_variables(
        self,
        var_names: np.ndarray,
        var_values: np.ndarray,
        stats: Optional[Dict] = None,
    ) -> None:
        variables = self._var_names_to_vars(var_names)
        for (var, val) in zip(variables, var_values):
            var.fix(val)
            self.solver.update_var(var)

    def optimize(self) -> None:
        if self.is_persistent:
            self.results = self.solver.solve(
                tee=True,
                warmstart=self._is_warm_start_available,
            )
        else:
            self.results = self.solver.solve(
                self.model,
                tee=True,
            )
            print(self.results)

    def relax(self) -> "AbstractModel":
        relaxed = self.model.clone()
        for var in relaxed.component_objects(Var):
            for idx in var:
                if var[idx].domain == pyomo.core.base.set_types.Binary:
                    lb, ub = var[idx].bounds
                    var[idx].setlb(lb)
                    var[idx].setub(ub)
                    var[idx].domain = pyomo.core.base.set_types.Reals
        return PyomoModel(relaxed, self.solver_name)

    def set_warm_starts(
        self,
        var_names: np.ndarray,
        var_values: np.ndarray,
        stats: Optional[Dict] = None,
    ) -> None:
        assert len(var_values.shape) == 2
        (n_starts, n_vars) = var_values.shape
        assert len(var_names.shape) == 1
        assert var_names.shape[0] == n_vars
        assert n_starts == 1, "Pyomo does not support multiple warm starts"
        variables = self._var_names_to_vars(var_names)
        for (var, val) in zip(variables, var_values[0, :]):
            var.value = val
        self._is_warm_start_available = True

    def _extract_after_load_vars(self, h5):
        names: List[str] = []
        types: List[str] = []
        upper_bounds: List[float] = []
        lower_bounds: List[float] = []
        obj_coeffs: List[float] = []

        obj = None
        obj_offset = 0.0
        obj_count = 0
        for obj in self.model.component_objects(Objective):
            obj, obj_offset = self._parse_pyomo_expr(obj.expr)
            obj_count += 1
        assert obj_count == 1, f"One objective function expected; found {obj_count}"

        for (i, var) in enumerate(self.model.component_objects(pyomo.core.Var)):
            for idx in var:
                v = var[idx]

                # Variable name
                if idx is None:
                    names.append(var.name)
                else:
                    names.append(var[idx].name)

                # Variable type
                if v.domain == pyomo.core.Binary:
                    types.append("B")
                elif v.domain in [
                    pyomo.core.Reals,
                    pyomo.core.NonNegativeReals,
                    pyomo.core.NonPositiveReals,
                    pyomo.core.NegativeReals,
                    pyomo.core.PositiveReals,
                ]:
                    types.append("C")
                else:
                    raise Exception(f"unknown variable domain: {v.domain}")

                # Variable upper/lower bounds
                lb, ub = v.bounds
                upper_bounds.append(float(ub))
                lower_bounds.append(float(lb))

                # Objective coefficients
                if v.name in obj:
                    obj_coeffs.append(obj[v.name])
                else:
                    obj_coeffs.append(0.0)

        h5.put_array("static_var_names", np.array(names, dtype="S"))
        h5.put_array("static_var_types", np.array(types, dtype="S"))
        h5.put_array("static_var_lower_bounds", np.array(lower_bounds))
        h5.put_array("static_var_upper_bounds", np.array(upper_bounds))
        h5.put_array("static_var_obj_coeffs", np.array(obj_coeffs))
        h5.put_scalar("static_obj_offset", obj_offset)

    def _extract_after_load_constrs(self, h5):
        names: List[str] = []
        rhs: List[float] = []
        senses: List[str] = []
        lhs_row: List[int] = []
        lhs_col: List[int] = []
        lhs_data: List[float] = []

        varname_to_idx = {}
        for var in self.model.component_objects(Var):
            for idx in var:
                varname = var.name
                if idx is not None:
                    varname = var[idx].name
                varname_to_idx[varname] = len(varname_to_idx)

        def _parse_constraint(c: pe.Constraint, row: int) -> None:
            # Extract RHS and sense
            has_ub = c.has_ub()
            has_lb = c.has_lb()
            assert (
                (not has_lb) or (not has_ub) or c.upper() == c.lower()
            ), "range constraints not supported"
            if not has_ub:
                senses.append(">")
                rhs.append(float(c.lower()))
            elif not has_lb:
                senses.append("<")
                rhs.append(float(c.upper()))
            else:
                senses.append("=")
                rhs.append(float(c.upper()))

            # Extract LHS
            expr = c.body
            if isinstance(expr, SumExpression):
                for term in expr._args_:
                    if isinstance(term, MonomialTermExpression):
                        lhs_row.append(row)
                        lhs_col.append(varname_to_idx[term._args_[1].name])
                        lhs_data.append(float(term._args_[0]))
                    elif isinstance(term, _GeneralVarData):
                        lhs_row.append(row)
                        lhs_col.append(varname_to_idx[term.name])
                        lhs_data.append(1.0)
                    else:
                        raise Exception(f"Unknown term type: {term.__class__.__name__}")
            elif isinstance(expr, _GeneralVarData):
                lhs_row.append(row)
                lhs_col.append(varname_to_idx[expr.name])
                lhs_data.append(1.0)
            else:
                raise Exception(f"Unknown expression type: {expr.__class__.__name__}")

        curr_row = 0
        for (i, constr) in enumerate(
            self.model.component_objects(pyomo.core.Constraint)
        ):
            if len(constr) > 0:
                for idx in constr:
                    names.append(constr[idx].name)
                    _parse_constraint(constr[idx], curr_row)
                    curr_row += 1
            else:
                names.append(constr.name)
                _parse_constraint(constr, curr_row)
                curr_row += 1

        lhs = coo_matrix((lhs_data, (lhs_row, lhs_col))).tocoo()
        h5.put_sparse("static_constr_lhs", lhs)
        h5.put_array("static_constr_names", np.array(names, dtype="S"))
        h5.put_array("static_constr_rhs", np.array(rhs))
        h5.put_array("static_constr_sense", np.array(senses, dtype="S"))

    def _extract_after_lp_vars(self, h5):
        rc = []
        values = []
        for var in self.model.component_objects(Var):
            for idx in var:
                v = var[idx]
                rc.append(self.model.rc[v])
                values.append(v.value)
        h5.put_array("lp_var_reduced_costs", np.array(rc))
        h5.put_array("lp_var_values", np.array(values))

    def _extract_after_lp_constrs(self, h5):
        dual = []
        slacks = []
        for constr in self.model.component_objects(pyomo.core.Constraint):
            for idx in constr:
                c = constr[idx]
                dual.append(self.model.dual[c])
                slacks.append(self.model.slack[c])
        h5.put_array("lp_constr_dual_values", np.array(dual))
        h5.put_array("lp_constr_slacks", np.array(slacks))

    def _extract_after_mip_vars(self, h5):
        values = []
        for var in self.model.component_objects(Var):
            for idx in var:
                v = var[idx]
                values.append(v.value)
        h5.put_array("mip_var_values", np.array(values))

    def _extract_after_mip_constrs(self, h5):
        slacks = []
        for constr in self.model.component_objects(pyomo.core.Constraint):
            for idx in constr:
                c = constr[idx]
                slacks.append(self.model.slack[c])
        h5.put_array("mip_constr_slacks", np.array(slacks))

    def _parse_pyomo_expr(self, expr: Any):
        lhs = {}
        offset = 0.0
        if isinstance(expr, SumExpression):
            for term in expr._args_:
                if isinstance(term, MonomialTermExpression):
                    lhs[term._args_[1].name] = float(term._args_[0])
                elif isinstance(term, _GeneralVarData):
                    lhs[term.name] = 1.0
                elif isinstance(term, Number):
                    offset += term
                else:
                    raise Exception(f"Unknown term type: {term.__class__.__name__}")
        elif isinstance(expr, _GeneralVarData):
            lhs[expr.name] = 1.0
        else:
            raise Exception(f"Unknown expression type: {expr.__class__.__name__}")
        return lhs, offset

    def _gap(self, zp, zd, tol=1e-6):
        # Reference: https://www.gurobi.com/documentation/9.5/refman/mipgap2.html
        if abs(zp) < tol:
            if abs(zd) < tol:
                return 0
            else:
                return float("inf")
        else:
            return abs(zp - zd) / abs(zp)

    def _get_sense(self):
        for obj in self.model.component_objects(Objective):
            sense = obj.sense
            if sense == pyomo.core.kernel.objective.minimize:
                return "min"
            elif sense == pyomo.core.kernel.objective.maximize:
                return "max"
            else:
                raise Exception(f"Unknown sense: ${sense}")

    def write(self, filename: str) -> None:
        self.model.write(filename, io_options={"symbolic_solver_labels": True})



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

import numpy as np
from miplearn.problems.tsp import (
    TravelingSalesmanData,
    TravelingSalesmanGenerator,
    build_tsp_model,
)
from scipy.spatial.distance import pdist, squareform
from scipy.stats import randint, uniform


def test_tsp_generator() -> None:
    np.random.seed(42)
    gen = TravelingSalesmanGenerator(
        x=uniform(loc=0.0, scale=1000.0),
        y=uniform(loc=0.0, scale=1000.0),
        n=randint(low=3, high=4),
        gamma=uniform(loc=1.0, scale=0.25),
        fix_cities=True,
        round=True,
    )
    data = gen.generate(2)
    assert data[0].distances.tolist() == [
        [0.0, 591.0, 996.0],
        [591.0, 0.0, 765.0],
        [996.0, 765.0, 0.0],
    ]
    assert data[1].distances.tolist() == [
        [0.0, 556.0, 853.0],
        [556.0, 0.0, 779.0],
        [853.0, 779.0, 0.0],
    ]


def test_tsp() -> None:
    data = TravelingSalesmanData(
        n_cities=6,
        distances=squareform(
            pdist(
                [
                    [0.0, 0.0],
                    [1.0, 0.0],
                    [2.0, 0.0],
                    [3.0, 0.0],
                    [0.0, 1.0],
                    [3.0, 1.0],
                ]
            )
        ),
    )
    model = build_tsp_model(data)
    model.optimize()
    assert model.inner.getAttr("x", model.inner.getVars()) == [
        1.0,
        0.0,
        0.0,
        1.0,
        0.0,
        1.0,
        0.0,
        0.0,
        0.0,
        1.0,
        0.0,
        0.0,
        0.0,
        1.0,
        1.0,
    ]



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

from dataclasses import dataclass
from typing import List, Tuple, Optional, Any

import gurobipy as gp
import networkx as nx
import numpy as np
from gurobipy import quicksum, GRB, tuplelist
from scipy.spatial.distance import pdist, squareform
from scipy.stats import uniform, randint
from scipy.stats.distributions import rv_frozen
import logging

from miplearn.solvers.gurobi import GurobiModel

logger = logging.getLogger(__name__)


@dataclass
class TravelingSalesmanData:
    n_cities: int
    distances: np.ndarray


class TravelingSalesmanGenerator:
    """Random generator for the Traveling Salesman Problem."""

    def __init__(
        self,
        x: rv_frozen = uniform(loc=0.0, scale=1000.0),
        y: rv_frozen = uniform(loc=0.0, scale=1000.0),
        n: rv_frozen = randint(low=100, high=101),
        gamma: rv_frozen = uniform(loc=1.0, scale=0.0),
        fix_cities: bool = True,
        round: bool = True,
    ) -> None:
        """Initializes the problem generator.

        Initially, the generator creates n cities (x_1,y_1),...,(x_n,y_n) where n,
        x_i and y_i are sampled independently from the provided probability
        distributions `n`, `x` and `y`. For each (unordered) pair of cities (i,j),
        the distance d[i,j] between them is set to:

            d[i,j] = gamma[i,j] \\sqrt{(x_i - x_j)^2 + (y_i - y_j)^2}

        where gamma is sampled from the provided probability distribution `gamma`.

        If fix_cities=True, the list of cities is kept the same for all generated
        instances. The gamma values, and therefore also the distances, are still
        different.

        By default, all distances d[i,j] are rounded to the nearest integer.  If
        `round=False` is provided, this rounding will be disabled.

        Arguments
        ---------
        x: rv_continuous
            Probability distribution for the x-coordinate of each city.
        y: rv_continuous
            Probability distribution for the y-coordinate of each city.
        n: rv_discrete
            Probability distribution for the number of cities.
        fix_cities: bool
            If False, cities will be resampled for every generated instance. Otherwise, list
            of cities will be computed once, during the constructor.
        round: bool
            If True, distances are rounded to the nearest integer.
        """
        assert isinstance(x, rv_frozen), "x should be a SciPy probability distribution"
        assert isinstance(y, rv_frozen), "y should be a SciPy probability distribution"
        assert isinstance(n, rv_frozen), "n should be a SciPy probability distribution"
        assert isinstance(
            gamma,
            rv_frozen,
        ), "gamma should be a SciPy probability distribution"
        self.x = x
        self.y = y
        self.n = n
        self.gamma = gamma
        self.round = round

        if fix_cities:
            self.fixed_n: Optional[int]
            self.fixed_cities: Optional[np.ndarray]
            self.fixed_n, self.fixed_cities = self._generate_cities()
        else:
            self.fixed_n = None
            self.fixed_cities = None

    def generate(self, n_samples: int) -> List[TravelingSalesmanData]:
        def _sample() -> TravelingSalesmanData:
            if self.fixed_cities is not None:
                assert self.fixed_n is not None
                n, cities = self.fixed_n, self.fixed_cities
            else:
                n, cities = self._generate_cities()
            distances = squareform(pdist(cities)) * self.gamma.rvs(size=(n, n))
            distances = np.tril(distances) + np.triu(distances.T, 1)
            if self.round:
                distances = distances.round()
            return TravelingSalesmanData(n, distances)

        return [_sample() for _ in range(n_samples)]

    def _generate_cities(self) -> Tuple[int, np.ndarray]:
        n = self.n.rvs()
        cities = np.array([(self.x.rvs(), self.y.rvs()) for _ in range(n)])
        return n, cities


def build_tsp_model(data: TravelingSalesmanData) -> GurobiModel:
    edges = tuplelist(
        (i, j) for i in range(data.n_cities) for j in range(i + 1, data.n_cities)
    )
    model = gp.Model()

    # Decision variables
    x = model.addVars(edges, vtype=GRB.BINARY, name="x")

    model._x = x
    model._edges = edges
    model._n_cities = data.n_cities

    # Objective function
    model.setObjective(quicksum(x[(i, j)] * data.distances[i, j] for (i, j) in edges))

    # Eq: Must choose two edges adjacent to each node
    model.addConstrs(
        (
            quicksum(x[min(i, j), max(i, j)] for j in range(data.n_cities) if i != j)
            == 2
            for i in range(data.n_cities)
        ),
        name="eq_degree",
    )

    def find_violations(model: GurobiModel) -> List[Any]:
        violations = []
        x = model.inner.cbGetSolution(model.inner._x)
        selected_edges = [e for e in model.inner._edges if x[e] > 0.5]
        graph = nx.Graph()
        graph.add_edges_from(selected_edges)
        for component in list(nx.connected_components(graph)):
            if len(component) < model.inner._n_cities:
                cut_edges = [
                    e
                    for e in model.inner._edges
                    if (e[0] in component and e[1] not in component)
                    or (e[0] not in component and e[1] in component)
                ]
                violations.append(cut_edges)
        return violations

    def fix_violations(model: GurobiModel, violations: List[Any], where: str) -> None:
        for violation in violations:
            constr = quicksum(model.inner._x[e[0], e[1]] for e in violation) >= 2
            if where == "cb":
                model.inner.cbLazy(constr)
            else:
                model.inner.addConstr(constr)
        logger.info(f"tsp: added {len(violations)} subtour elimination constraints")

    model.update()

    return GurobiModel(
        model,
        find_violations=find_violations,
        fix_violations=fix_violations,
    )



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

from dataclasses import dataclass
from typing import List, Optional

import gurobipy as gp
import numpy as np
from gurobipy import quicksum, GRB
from scipy.spatial.distance import pdist, squareform
from scipy.stats import uniform, randint
from scipy.stats.distributions import rv_frozen

from miplearn.solvers.gurobi import GurobiModel


@dataclass
class PMedianData:
    """Data for the capacitated p-median problem

    Args
    ----
    distances
        Matrix of distances between customer i and facility j.
    demands
        Customer demands.
    p
        Number of medians that need to be chosen.
    capacities
        Facility capacities.
    """

    distances: np.ndarray
    demands: np.ndarray
    p: int
    capacities: np.ndarray


class PMedianGenerator:
    """Random generator for the capacitated p-median problem.

    This class first decides the number of customers and the parameter `p` by
    sampling the provided `n` and `p` distributions, respectively. Then, for each
    customer `i`, the class builds its geographical location `(xi, yi)` by sampling
    the provided `x` and `y` distributions. For each `i`, the demand for customer `i`
    and the capacity of facility `i` are decided by sampling the distributions
    `demands` and `capacities`, respectively. Finally, the costs `w[i,j]` are set to
    the Euclidean distance between the locations of customers `i` and `j`.

    If `fixed=True`, then the number of customers, their locations, the parameter
    `p`, the demands and the capacities are only sampled from their respective
    distributions exactly once, to build a reference instance which is then
    perturbed. Specifically, for each perturbation, the distances, demands and
    capacities are multiplied by factors sampled from the distributions
    `distances_jitter`, `demands_jitter` and `capacities_jitter`, respectively. The
    result is a list of instances that have the same set of customers, but slightly
    different demands, capacities and distances.

    Parameters
    ----------
    x
        Probability distribution for the x-coordinate of the points.
    y
        Probability distribution for the y-coordinate of the points.
    n
        Probability distribution for the number of customer.
    p
        Probability distribution for the number of medians.
    demands
        Probability distribution for the customer demands.
    capacities
        Probability distribution for the facility capacities.
    distances_jitter
        Probability distribution for the random scaling factor applied to distances.
    demands_jitter
        Probability distribution for the random scaling factor applied to demands.
    capacities_jitter
        Probability distribution for the random scaling factor applied to capacities.
    fixed
        If `True`, then customer are kept the same across instances.
    """

    def __init__(
        self,
        x: rv_frozen = uniform(loc=0.0, scale=100.0),
        y: rv_frozen = uniform(loc=0.0, scale=100.0),
        n: rv_frozen = randint(low=100, high=101),
        p: rv_frozen = randint(low=10, high=11),
        demands: rv_frozen = uniform(loc=0, scale=20),
        capacities: rv_frozen = uniform(loc=0, scale=100),
        distances_jitter: rv_frozen = uniform(loc=1.0, scale=0.0),
        demands_jitter: rv_frozen = uniform(loc=1.0, scale=0.0),
        capacities_jitter: rv_frozen = uniform(loc=1.0, scale=0.0),
        fixed: bool = True,
    ):
        self.x = x
        self.y = y
        self.n = n
        self.p = p
        self.demands = demands
        self.capacities = capacities
        self.distances_jitter = distances_jitter
        self.demands_jitter = demands_jitter
        self.capacities_jitter = capacities_jitter
        self.fixed = fixed
        self.ref_data: Optional[PMedianData] = None

    def generate(self, n_samples: int) -> List[PMedianData]:
        def _sample() -> PMedianData:
            if self.ref_data is None:
                n = self.n.rvs()
                p = self.p.rvs()
                loc = np.array([(self.x.rvs(), self.y.rvs()) for _ in range(n)])
                distances = squareform(pdist(loc))
                demands = self.demands.rvs(n)
                capacities = self.capacities.rvs(n)
            else:
                n = self.ref_data.demands.shape[0]
                distances = self.ref_data.distances * self.distances_jitter.rvs(
                    size=(n, n)
                )
                distances = np.tril(distances) + np.triu(distances.T, 1)
                demands = self.ref_data.demands * self.demands_jitter.rvs(n)
                capacities = self.ref_data.capacities * self.capacities_jitter.rvs(n)
                p = self.ref_data.p

            data = PMedianData(
                distances=distances.round(2),
                demands=demands.round(2),
                p=p,
                capacities=capacities.round(2),
            )

            if self.fixed and self.ref_data is None:
                self.ref_data = data

            return data

        return [_sample() for _ in range(n_samples)]


def build_pmedian_model(data: PMedianData) -> GurobiModel:
    """Converts capacitated p-median data into a concrete Gurobipy model."""
    model = gp.Model()
    n = len(data.demands)

    # Decision variables
    x = model.addVars(n, n, vtype=GRB.BINARY, name="x")
    y = model.addVars(n, vtype=GRB.BINARY, name="y")

    # Objective function
    model.setObjective(
        quicksum(data.distances[i, j] * x[i, j] for i in range(n) for j in range(n))
    )

    # Eq: Must serve each customer
    model.addConstrs(
        (quicksum(x[i, j] for j in range(n)) == 1 for i in range(n)),
        name="eq_demand",
    )

    # Eq: Must choose p medians
    model.addConstr(
        quicksum(y[j] for j in range(n)) == data.p,
        name="eq_choose",
    )

    # Eq: Must not exceed capacity
    model.addConstrs(
        (
            quicksum(data.demands[i] * x[i, j] for i in range(n))
            <= data.capacities[j] * y[j]
            for j in range(n)
        ),
        name="eq_capacity",
    )

    model.update()
    return GurobiModel(model)



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.
import logging
import random

from miplearn.parallel import p_umap

from config import challenges, n_jobs, collectors

logger = logging.getLogger()


def collect():
    def _collect(args):
        challenge_idx, sample_idx = args
        challenge = challenges[challenge_idx]
        data = f"data/{challenge.name}/{challenge.name}-{sample_idx:05d}.pkl.gz"
        try:
            for c in collectors:
                c.collect([data], challenge.build_model)
        except:
            logger.exception(f"Failed sample: {challenge.name} {sample_idx}")

    combinations = [
        (challenge_idx, sample_idx)
        for (challenge_idx, challenge) in enumerate(challenges)
        for sample_idx in range(challenge.n_samples)
    ]

    random.shuffle(combinations)

    if n_jobs > 1:
        p_umap(
            _collect,
            combinations,
            num_cpus=n_jobs,
            desc="collect",
            smoothing=0,
        )
    else:
        for args in combinations:
            _collect(args)


if __name__ == "__main__":
    collect()



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.
from tempfile import NamedTemporaryFile
from typing import Any

import numpy as np
from scipy.sparse import coo_matrix

from miplearn.h5 import H5File


def test_h5() -> None:
    file = NamedTemporaryFile()
    h5 = H5File(file.name)
    _assert_roundtrip_scalar(h5, "A")
    _assert_roundtrip_scalar(h5, True)
    _assert_roundtrip_scalar(h5, 1)
    _assert_roundtrip_scalar(h5, 1.0)
    assert h5.get_scalar("unknown-key") is None

    _assert_roundtrip_array(h5, np.array([True, False]))
    _assert_roundtrip_array(h5, np.array([1, 2, 3]))
    _assert_roundtrip_array(h5, np.array([1.0, 2.0, 3.0]))
    _assert_roundtrip_array(h5, np.array(["A", "BB", "CCC"], dtype="S"))
    assert h5.get_array("unknown-key") is None

    _assert_roundtrip_sparse(
        h5,
        coo_matrix(
            [
                [1.0, 0.0, 0.0],
                [0.0, 2.0, 3.0],
                [0.0, 0.0, 4.0],
            ],
        ),
    )
    assert h5.get_sparse("unknown-key") is None


def _assert_roundtrip_array(h5: H5File, original: np.ndarray) -> None:
    h5.put_array("key", original)
    recovered = h5.get_array("key")
    assert recovered is not None
    assert isinstance(recovered, np.ndarray)
    assert (recovered == original).all()


def _assert_roundtrip_scalar(h5: H5File, original: Any) -> None:
    h5.put_scalar("key", original)
    recovered = h5.get_scalar("key")
    assert recovered == original
    assert recovered is not None
    assert isinstance(
        recovered, original.__class__
    ), f"Expected {original.__class__}, found {recovered.__class__} instead"


def _assert_roundtrip_sparse(h5: H5File, original: coo_matrix) -> None:
    h5.put_sparse("key", original)
    recovered = h5.get_sparse("key")
    assert recovered is not None
    assert isinstance(recovered, coo_matrix)
    assert (original != recovered).sum() == 0



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

import numpy as np

from miplearn.problems.setpack import (
    SetPackData,
    build_setpack_model,
)


def test_setpack() -> None:
    data = SetPackData(
        costs=np.array([5, 10, 12, 6, 8]),
        incidence_matrix=np.array(
            [
                [1, 0, 0, 1, 0],
                [1, 1, 0, 0, 0],
                [0, 0, 1, 1, 1],
            ],
        ),
    )
    model = build_setpack_model(data)
    model.optimize()
    assert model.inner.objval == -22.0



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.
from typing import List, Dict, Any
from unittest.mock import Mock

from miplearn.components.primal.actions import SetWarmStart, FixVariables
from miplearn.components.primal.expert import ExpertPrimalComponent


def test_expert(multiknapsack_h5: List[str]) -> None:
    model = Mock()
    stats: Dict[str, Any] = {}
    comp = ExpertPrimalComponent(action=SetWarmStart())
    comp.before_mip(multiknapsack_h5[0], model, stats)
    model.set_warm_starts.assert_called()
    names, starts, _ = model.set_warm_starts.call_args.args
    assert names.shape == (100,)
    assert starts.shape == (1, 100)

    comp = ExpertPrimalComponent(action=FixVariables())
    comp.before_mip(multiknapsack_h5[0], model, stats)
    model.fix_variables.assert_called()
    names, v, _ = model.fix_variables.call_args.args
    assert names.shape == (100,)
    assert v.shape == (100,)



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

from dataclasses import dataclass
from typing import List

import gurobipy as gp
import numpy as np
import pyomo.environ as pe
from gurobipy.gurobipy import GRB
from scipy.stats import uniform, randint
from scipy.stats.distributions import rv_frozen

from miplearn.solvers.gurobi import GurobiModel
from miplearn.solvers.pyomo import PyomoModel


@dataclass
class SetCoverData:
    costs: np.ndarray
    incidence_matrix: np.ndarray


class SetCoverGenerator:
    def __init__(
        self,
        n_elements: rv_frozen = randint(low=50, high=51),
        n_sets: rv_frozen = randint(low=100, high=101),
        costs: rv_frozen = uniform(loc=0.0, scale=100.0),
        costs_jitter: rv_frozen = uniform(loc=-5.0, scale=10.0),
        K: rv_frozen = uniform(loc=25.0, scale=0.0),
        density: rv_frozen = uniform(loc=0.02, scale=0.00),
        fix_sets: bool = True,
    ):
        self.n_elements = n_elements
        self.n_sets = n_sets
        self.costs = costs
        self.costs_jitter = costs_jitter
        self.density = density
        self.K = K
        self.fix_sets = fix_sets
        self.fixed_costs = None
        self.fixed_matrix = None

    def generate(self, n_samples: int) -> List[SetCoverData]:
        def _sample() -> SetCoverData:
            if self.fixed_matrix is None:
                n_sets = self.n_sets.rvs()
                n_elements = self.n_elements.rvs()
                density = self.density.rvs()

                incidence_matrix = np.random.rand(n_elements, n_sets) < density
                incidence_matrix = incidence_matrix.astype(int)

                # Ensure each element belongs to at least one set
                for j in range(n_elements):
                    if incidence_matrix[j, :].sum() == 0:
                        incidence_matrix[j, randint(low=0, high=n_sets).rvs()] = 1

                # Ensure each set contains at least one element
                for i in range(n_sets):
                    if incidence_matrix[:, i].sum() == 0:
                        incidence_matrix[randint(low=0, high=n_elements).rvs(), i] = 1

                costs = self.costs.rvs(n_sets) + self.K.rvs() * incidence_matrix.sum(
                    axis=0
                )
                if self.fix_sets:
                    self.fixed_matrix = incidence_matrix
                    self.fixed_costs = costs
            else:
                incidence_matrix = self.fixed_matrix
                (_, n_sets) = incidence_matrix.shape
                costs = self.fixed_costs * self.costs_jitter.rvs(n_sets)
            return SetCoverData(
                costs=costs.round(2),
                incidence_matrix=incidence_matrix,
            )

        return [_sample() for _ in range(n_samples)]


def build_setcover_model_gurobipy(data: SetCoverData) -> GurobiModel:
    (n_elements, n_sets) = data.incidence_matrix.shape
    model = gp.Model()
    x = model.addMVar(n_sets, vtype=GRB.BINARY, name="x")
    model.addConstr(data.incidence_matrix @ x >= np.ones(n_elements), name="eqs")
    model.setObjective(data.costs @ x)
    model.update()
    return GurobiModel(model)


def build_setcover_model_pyomo(
    data: SetCoverData,
    solver="gurobi_persistent",
) -> PyomoModel:
    (n_elements, n_sets) = data.incidence_matrix.shape
    model = pe.ConcreteModel()
    model.sets = pe.Set(initialize=range(n_sets))
    model.x = pe.Var(model.sets, domain=pe.Boolean, name="x")
    model.eqs = pe.Constraint(pe.Any)
    for i in range(n_elements):
        model.eqs[i] = (
            sum(data.incidence_matrix[i, j] * model.x[j] for j in range(n_sets)) >= 1
        )
    model.obj = pe.Objective(
        expr=sum(data.costs[j] * model.x[j] for j in range(n_sets))
    )
    return PyomoModel(model, solver)



# Modified version of: https://github.com/swansonk14/p_tqdm
# Copyright (c) 2022 Kyle Swanson
# MIT License

from collections.abc import Sized
from typing import Any, Callable, Generator, Iterable, List

from pathos.multiprocessing import _ProcessPool as Pool
from tqdm.auto import tqdm


def _parallel(function: Callable, *iterables: Iterable, **kwargs: Any) -> Generator:
    # Determine length of tqdm (equal to length of the shortest iterable or total kwarg)
    total = kwargs.pop("total", None)
    lengths = [len(iterable) for iterable in iterables if isinstance(iterable, Sized)]
    length = total or (min(lengths) if lengths else None)

    # Create parallel generator
    num_cpus = kwargs.pop("num_cpus", 1)
    maxtasksperchild = kwargs.pop("maxtasksperchild", 1)
    chunksize = kwargs.pop("chunksize", 1)
    with Pool(num_cpus, maxtasksperchild=maxtasksperchild) as pool:
        for item in tqdm(
            pool.imap_unordered(function, *iterables, chunksize=chunksize),
            total=length,
            **kwargs
        ):
            yield item


def p_umap(function: Callable, *iterables: Iterable, **kwargs: Any) -> List[Any]:
    return list(_parallel(function, *iterables, **kwargs))



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.
from typing import Optional, List

import numpy as np

from miplearn.extractors.abstract import FeaturesExtractor
from miplearn.h5 import H5File


class H5FieldsExtractor(FeaturesExtractor):
    def __init__(
        self,
        instance_fields: Optional[List[str]] = None,
        var_fields: Optional[List[str]] = None,
        constr_fields: Optional[List[str]] = None,
    ):
        self.instance_fields = instance_fields
        self.var_fields = var_fields
        self.constr_fields = constr_fields

    def get_instance_features(self, h5: H5File) -> np.ndarray:
        if self.instance_fields is None:
            raise Exception("No instance fields provided")
        x = []
        for field in self.instance_fields:
            try:
                data = h5.get_array(field)
            except ValueError:
                data = h5.get_scalar(field)
            assert data is not None
            x.append(data)
        x = np.hstack(x)
        assert len(x.shape) == 1
        return x

    def get_var_features(self, h5: H5File) -> np.ndarray:
        var_types = h5.get_array("static_var_types")
        assert var_types is not None
        n_vars = len(var_types)
        if self.var_fields is None:
            raise Exception("No var fields provided")
        return self._extract(h5, self.var_fields, n_vars)

    def get_constr_features(self, h5: H5File) -> np.ndarray:
        constr_sense = h5.get_array("static_constr_sense")
        assert constr_sense is not None
        n_constr = len(constr_sense)
        if self.constr_fields is None:
            raise Exception("No constr fields provided")
        return self._extract(h5, self.constr_fields, n_constr)

    def _extract(self, h5, fields, n_expected):
        x = []
        for field in fields:
            try:
                data = h5.get_array(field)
            except ValueError:
                v = h5.get_scalar(field)
                data = np.repeat(v, n_expected)
            assert data is not None
            assert len(data.shape) == 1
            assert data.shape[0] == n_expected
            x.append(data)
        features = np.vstack(x).T
        assert len(features.shape) == 2
        assert features.shape[0] == n_expected
        return features



#  MIPLearn: Extensible Framework for Learning-Enhanced Mixed-Integer Optimization
#  Copyright (C) 2020-2022, UChicago Argonne, LLC. All rights reserved.
#  Released under the modified BSD license. See COPYING.md for more details.

import logging
import random
import sys
from distutils.dir_util import mkpath
from glob import glob
from os.path import dirname, exists
from pathlib import Path
from tempfile import NamedTemporaryFile
from time import time
import json
from contextlib import redirect_stdout
import os
from io import StringIO
import re
import pickle

import pandas as pd

from miplearn.h5 import H5File
from miplearn.io import load, _RedirectOutput
from miplearn.parallel import p_umap

from config import challenges, methods, n_jobs, timelimit_benchmark, n_seeds

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
stdout_handler = logging.StreamHandler(sys.stdout)
stdout_handler.setLevel(logging.ERROR)
stdout_handler.setFormatter(logging.Formatter("[%(asctime)s] %(message)s"))
logger.addHandler(stdout_handler)

# -----------------------------------------------------------------------------


def fit():
    def _fit(args):
        # Parse arguments
        method_idx, challenge_idx = args
        method = methods[method_idx]
        challenge = challenges[challenge_idx]

        # Compute filenames
        pkl_filename = f"results/{challenge.name}/{method.name}/model.pkl"
        log_filename = pkl_filename.replace(".pkl", ".log")

        # Skip instances that have already been processed
        if exists(pkl_filename):
            return

        # Create folder
        mkpath(dirname(log_filename))

        # Set up logger
        file_handler = logging.FileHandler(log_filename, mode="w")
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(logging.Formatter("[%(asctime)s] %(message)s"))
        logger.addHandler(file_handler)

        # Fit
        data = sorted(glob(f"data/{challenge.name}/*.pkl.gz"))
        h5 = [f.replace(".pkl.gz", ".h5") for f in data]
        train_h5 = h5[: challenge.n_train]
        for c in method.components:
            c.fit(train_h5)

        # Pickle
        _write_pkl(method, pkl_filename)

        # Tear down logger
        logger.removeHandler(file_handler)

    initial_time = time()
    combinations = [
        (method_idx, challenge_idx)
        for method_idx in range(len(methods))
        for challenge_idx in range(len(challenges))
    ]
    random.shuffle(combinations)
    if n_jobs > 1:
        p_umap(
            _fit,
            combinations,
            num_cpus=n_jobs,
            desc="fit",
            smoothing=0,
        )
    else:
        [_fit(c) for c in combinations]

    elapsed_time = time() - initial_time
    print(f"total time: {elapsed_time:.2f} s")


# -----------------------------------------------------------------------------


def benchmark():
    def _benchmark(args):
        # Parse arguments
        challenge_idx, sample_idx, method_idx, mip_seed = args
        challenge = challenges[challenge_idx]
        method = methods[method_idx]

        # Compute filenames
        log_filename = f"results/{challenge.name}/{method.name}/{sample_idx:05d}-{mip_seed:05d}.log"
        stats_filename = log_filename.replace(".log", ".json")
        pkl_filename = f"results/{challenge.name}/{method.name}/model.pkl"
        data_filename = (
            f"data/{challenge.name}/{challenge.name}-{sample_idx:05d}.pkl.gz"
        )
        h5_filename = f"data/{challenge.name}/{challenge.name}-{sample_idx:05d}.h5"

        # Skip instances that have already been processed
        if exists(stats_filename):
            return

        # Create folder
        mkpath(dirname(log_filename))

        # Set up logger
        file_handler = logging.FileHandler(log_filename, mode="w")
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(logging.Formatter("[%(asctime)s] %(message)s"))
        logger.addHandler(file_handler)

        try:
            with NamedTemporaryFile() as tempfile:
                with H5File(tempfile.name) as h5:
                    components = _read_pkl(pkl_filename).components
                    with _RedirectOutput([]):
                        model = load(data_filename, challenge.build_model)
                    model.extract_after_load(h5)
                    stats = {}
                    for comp in components:
                        comp.before_mip(h5_filename, model, stats)
                    streams = [StringIO()]
                    with _RedirectOutput(streams):
                        model.optimize()
                    log = streams[0].getvalue()
                    model.extract_after_mip(h5)

                    var_types = h5.get_array("static_var_types")
                    constr_lhs = h5.get_sparse("static_constr_lhs")
                    n_vars = len(var_types)
                    n_int_vars = int((var_types == b"I").sum())
                    n_bin_vars = int((var_types == b"B").sum())
                    mip_val = h5.get_scalar("mip_obj_value")
                    mip_bound = h5.get_scalar("mip_obj_bound")
                    mip_gap = h5.get_scalar("mip_gap")
                    mip_time = h5.get_scalar("mip_wallclock_time")
                    mip_nodes = h5.get_scalar("mip_node_count")
                    (ws_value, ws_time), ws_gap = (
                        _extract_warm_start_stats(log),
                        None,
                    )
                    if ws_value is not None:
                        ws_gap = 100 * abs(ws_value - mip_bound) / abs(ws_value)
                    if "Heuristic" in stats:
                        mip_bound = None
                        mip_gap = None
                    stats.update(
                        {
                            "Challenge": challenge.name,
                            "Instance": f"{challenge.name}-{sample_idx:05d}",
                            "Method": method.name,
                            "MIP seed": mip_seed,
                            "Objective value": mip_val,
                            "Objective bound": mip_bound,
                            "Wallclock time (s)": mip_time,
                            "Node count": mip_nodes,
                            "Relative MIP gap (%)": mip_gap,
                            "Variables": n_vars,
                            "Integer variables": n_int_vars,
                            "Binary variables": n_bin_vars,
                            "Constraints": constr_lhs.shape[0],
                            "Non-zeros": constr_lhs.nnz,
                            "WS: Objective value": ws_value,
                            "WS: Processing time (s)": ws_time,
                            "WS: Relative MIP gap (%)": ws_gap,
                        }
                    )

                    # Write stats
                    _write_json(stats, stats_filename)
        except:
            logger.exception(f"Failed sample: {args}")

        # Tear down logging
        logger.removeHandler(file_handler)

    initial_time = time()
    combinations = [
        (challenge_idx, sample_idx, method_idx, mip_seed)
        for (challenge_idx, challenge) in enumerate(challenges)
        for sample_idx in range(challenge.n_train, challenge.n_samples)
        for mip_seed in [1000 + i for i in range(n_seeds)]
        for (method_idx, method) in enumerate(methods)
    ]
    random.shuffle(combinations)

    if n_jobs > 1:
        p_umap(
            _benchmark,
            combinations,
            num_cpus=n_jobs,
            desc="benchmark",
            smoothing=0,
        )
    else:
        [_benchmark(c) for c in combinations]

    elapsed_time = time() - initial_time
    print(f"total time: {elapsed_time:.2f} s")


# -----------------------------------------------------------------------------


def _write_csv(results, filename):
    Path(dirname(filename)).mkdir(parents=True, exist_ok=True)
    df = pd.DataFrame(results)
    df["Method"] = pd.Categorical(df["Method"], [m.name for m in methods])
    df.sort_values(by=["Challenge", "Instance", "Method", "MIP seed"], inplace=True)
    df.to_csv(filename, index=False)


def _write_json(obj, filename):
    with open(filename, "w") as f:
        json.dump(obj, f, indent=2)


def _write_pkl(obj, filename):
    with open(filename, "wb") as f:
        pickle.dump(obj, f)


def _read_pkl(filename):
    with open(filename, "rb") as f:
        return pickle.load(f)


def _silence_stdout():
    return redirect_stdout(open(os.devnull, "w"))


def _extract_warm_start_stats(log):
    ws_value = None
    ws_time = None
    for line in log.splitlines():
        matches = re.findall("Loaded user MIP start .*with objective (.*)", line)
        if len(matches) > 0:
            ws_value = float(matches[0])

        matches = re.findall("Processed .*MIP starts? in (.*) seconds", line)
        if len(matches) > 0:
            ws_time = float(matches[0])

    return ws_value, ws_time


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    fit()
    benchmark()
